import logging
import sys
from pathlib import Path
from typing import Optional, Union


class LoggerManager:
    """A modular and fully-featured logger class for managing logging across modules.

    This class sets up a logger with the specified name, log file, and logging level. It supports logging to both stdout and a file, with customizable formats and date formats.

    Example:
        >>> logger_manager = LoggerManager(name="my_module", prefix="my_prefix", debug=True)
        >>> logger = logger_manager.get_logger("my_module")
        >>>
        >>> logger.info("This is an info message.")
        >>> logger.debug("This is a debug message.")
        >>> logger.warning("This is a warning message.")
        >>> logger.error("This is an error message.")
        >>> logger.critical("This is a critical message.")
        >>> logger_manager.set_level(logging.WARNING)
        >>> logger_manager.remove_handler(logger)

    Attributes:
        name (str): The name of the logger.
        logger (logging.Logger): The configured logger instance.

    Methods:
        get_logger(): Returns the logger instance.
        set_level(level): Sets the logging level.
        add_handler(handler): Adds a logging handler.
        remove_handler(handler): Removes a logging handler.
    """

    def __init__(
        self,
        name: str,
        prefix: Optional[str] = "",
        debug: bool = False,
        verbose: bool = True,
        log_file: Optional[Union[str, Path]] = None,
        level: Optional[int] = None,
        to_console: bool = True,
        to_file: bool = True,
        log_format: str = "%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s",
        date_format: str = "%Y-%m-%d %H:%M:%S",
    ):
        """
        Initializes the LoggerManager.

        This class sets up a logger with the specified name, log file, and logging level. It supports logging to both stdout and a file, with customizable formats and date formats. The default log format is "%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s", and the default date format is "%Y-%m-%d %H:%M:%S".

        Args:
            name (str): Name of the logger.
            prefix (str, optional): Prefix for the log file directory.
            debug (bool, optional): If True, sets log level to DEBUG.
            verbose (bool, optional): If False, suppresses INFO and WARNING messages.
            log_file (str or Path, optional): Path to the log file.
            level (int, optional): Explicit logging level (overrides debug and verbose).
            to_console (bool, optional): Whether to log to console. Defaults to True.
            to_file (bool, optional): Whether to log to file. Defaults to True.
            log_format (str, optional): Format of the log messages.
            date_format (str, optional): Format of the date in log messages.

        Note:
            If both ``to_console`` and ``to_file`` are set to ``False``, the logger will not output any logs.

            If both ``to_console`` and ``to_file`` are set to ``True``, the logger will output logs to both stdout and a file.

            If ``to_console`` is set to ``True`` and ``to_file`` is set to False, the logger will output logs to stdout only.

            If ``to_console`` is set to False and to_file is set to ``True``, the logger will output logs to a file only.

            If no ``log_file`` is provided, the log file will be saved in the 'logs' directory with the name of the logger.

            If a ``prefix`` is provided, the log file will be saved in a subdirectory of 'logs' with the ``prefix`` as the directory name.

            If no ``prefix`` is provided, the log file will be saved in the 'logs'.

            The default log format is "%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s".

            The default date format is "%Y-%m-%d %H:%M:%S".

            The default logging level is ``logging.INFO``.

            The default values for ``to_console`` and ``to_file`` are True.
        """
        self.name = name
        self.logger = logging.getLogger(name)

        # Set default log level to INFO
        self.logger.setLevel(logging.INFO)

        # Determine log level
        if level is not None:
            self.logger.setLevel(level)
        elif debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        formatter = logging.Formatter(fmt=log_format, datefmt=date_format)

        # Check if handlers already exist to prevent duplicates
        if not self.logger.handlers:
            handlers = []

            if to_console:
                # Add StreamHandler to stdout
                stream_handler = logging.StreamHandler(sys.stdout)
                stream_handler.setFormatter(formatter)
                handlers.append(stream_handler)

            if to_file:
                # Determine log_file path
                if log_file:
                    log_file = Path(log_file)
                elif prefix:
                    # Construct log_file path using prefix
                    log_file = Path(f"{prefix}_output") / "logs" / f"{name}.log"
                else:
                    log_file = Path("logs") / f"{name}.log"

                # Ensure parent directories exist
                log_file.parent.mkdir(exist_ok=True, parents=True)

                file_handler = logging.FileHandler(log_file)
                file_handler.setFormatter(formatter)
                handlers.append(file_handler)

            # Apply verbose filtering if necessary
            for handler in handlers:
                if not verbose:
                    # Set handler level to ERROR to suppress INFO and WARNING
                    handler.setLevel(logging.ERROR)
                else:
                    # Ensure handler level is not set (inherits logger level)
                    handler.setLevel(logging.NOTSET)
                self.logger.addHandler(handler)

    def get_logger(self) -> logging.Logger:
        """Returns the logger instance.

        Returns:
            logging.Logger: The configured logger.
        """
        return self.logger

    def set_level(self, level: int) -> None:
        """Sets the logging level.

        Args:
            level (int): The new logging level.
        """
        self.logger.setLevel(level)

    def add_handler(self, handler: logging.Handler) -> None:
        """Adds a logging handler if it's not already added.

        Args:
            handler (logging.Handler): The handler to add.
        """
        if handler not in self.logger.handlers:
            self.logger.addHandler(handler)

    def remove_handler(self, handler: logging.Handler) -> None:
        """Removes a logging handler if it exists.

        Args:
            handler (logging.Handler): The handler to remove.
        """
        if handler in self.logger.handlers:
            self.logger.removeHandler(handler)



from pathlib import Path
from typing import Any, List, Optional

import numpy as np

from snpio.read_input.genotype_data import GenotypeData
from snpio.utils.custom_exceptions import (
    AlignmentError,
    AlignmentFileNotFoundError,
    AlignmentFormatError,
    PhylipAlignmentSampleMismatch,
    SequenceLengthError,
)
from snpio.utils.logging import LoggerManager


class PhylipReader(GenotypeData):
    """Class to read and write PHYLIP files.

    This class provides methods to read and write PHYLIP files. The PHYLIP format is a simple text format for representing multiple sequence alignments. The first line of a PHYLIP file contains the number of samples and the number of loci. Each subsequent line contains the sample ID followed by the sequence data. The sequence data can be in any format, but it is typically a string of nucleotides or amino acids.

    Example:
        >>> from snpio import PhylipReader
        >>>
        >>> phylip = PhylipReader(filename="example.phy", popmapfile="example.popmap", verbose=True)
        >>>
        >>> genotype_data.snp_data
        array([["A", "T", "T", "A"], ["C", "G", "G", "C"], ["A", "T", "T", "A"]], dtype="<U1")
        >>>
        >>> genotype_data.samples
        ["Sample1", "Sample2", "Sample3", "Sample4"]
        >>>
        >>> genotype_data.populations
        ["Pop1", "Pop1", "Pop2", "Pop2"]
        >>>
        >>> genotype_data.num_snps
        3
        >>>
        >>> genotype_data.num_inds
        4
        >>>
        >>> genotype_data.popmap
        >>> {"Sample1": "Pop1", "Sample2": "Pop1", "Sample3": "Pop2", "Sample4": "Pop2"}
        >>>
        >>> genotype_data.popmap_inverse
        {"Pop1": ["Sample1", "Sample2"], "Pop2": ["Sample3", "Sample4"]}
        >>>
        >>> genotype_data.ref
        ["A", "C", "A"]
        >>>
        >>> genotype_data.alt
        ["T", "G", "T"]
        >>>
        >>> genotype_data.missingness_reports()
        >>>
        >>> genotype_data.run_pca()
        >>>
        >>> genotype_data.write_phylip("output.str")

    Attributes:
        filename (str): Name of the PHYLIP file.
        popmapfile (str): Name of the population map file.
        force_popmap (bool): If True, the population map file is required.
        exclude_pops (List[str]): List of populations to exclude.
        include_pops (List[str]): List of populations to include.
        guidetree (str): Name of the guide tree file.
        qmatrix_iqtree (str): Name of the Q-matrix file generated by IQ-TREE.
        qmatrix (str): Name of the Q-matrix file.
        siterates (str): Name of the site rates file.
        siterates_iqtree (str): Name of the site rates file generated by IQ-TREE.
        plot_format (str): Format for saving plots. Default is 'png'.
        prefix (str): Prefix for output files.
        verbose (bool): If True, status updates are printed.
        samples (List[str]): List of sample IDs.
        snp_data (List[List[str]]): List of SNP data.
        num_inds (int): Number of individuals.
        num_snps (int): Number of SNPs.
        logger (Logger): Logger instance.
        debug (bool): If True, debug messages are printed.
    """

    def __init__(
        self,
        filename: Optional[str] = None,
        popmapfile: Optional[str] = None,
        force_popmap: bool = False,
        exclude_pops: Optional[List[str]] = None,
        include_pops: Optional[List[str]] = None,
        guidetree: Optional[str] = None,
        qmatrix_iqtree: Optional[str] = None,
        qmatrix: Optional[str] = None,
        siterates: Optional[str] = None,
        siterates_iqtree: Optional[str] = None,
        plot_format: Optional[str] = "png",
        prefix: str = "snpio",
        verbose: bool = False,
        debug: bool = False,
    ) -> None:
        """Initialize the PhylipReader class.

        This class inherits from the GenotypeData class and provides methods to read and write PHYLIP files. The PHYLIP format is a simple text format for representing multiple sequence alignments. The first line of a PHYLIP file contains the number of samples and the number of loci. Each subsequent line contains the sample ID followed by the sequence data. The sequence data can be in any format, but it is typically a string of nucleotides or amino acids.

        Args:
            filename (str, optional): Name of the PHYLIP file. Defaults to None.
            popmapfile (str, optional): Name of the population map file. Defaults to None.
            force_popmap (bool, optional): If True, the population map file is required. Defaults to False.
            exclude_pops (List[str], optional): List of populations to exclude. Defaults to None.
            include_pops (List[str], optional): List of populations to include. Defaults to None.
            guidetree (str, optional): Name of the guide tree file. Defaults to None.
            qmatrix_iqtree (str, optional): Name of the Q-matrix file generated by IQ-TREE. Defaults to None.
            qmatrix (str, optional): Name of the Q-matrix file. Defaults to None.
            siterates (str, optional): Name of the site rates file. Defaults to None.
            siterates_iqtree (str, optional): Name of the site rates file generated by IQ-TREE. Defaults to None.
            plot_format (str, optional): Format for saving plots. Default is 'png'. Defaults to 'png'.
            prefix (str, optional): Prefix for output files. Defaults to 'snpio'.
            verbose (bool, optional): If True, status updates are printed. Defaults to False.
            debug (bool, optional): If True, debug messages are printed. Defaults to False.

        Raises:
            AlignmentFormatError: If the PHYLIP file has an invalid format.
            AlignmentFileNotFoundError: If the PHYLIP file is not found.

        Note:
            The PHYLIP file format is a simple text format for representing multiple sequence alignments.

            The first line of a PHYLIP file contains the number of samples and the number of loci.

            Each subsequent line contains the sample ID followed by the sequence data.

            The sequence data can be in any format, but it is typically a string of nucleotides or amino acids.

            The PHYLIP file must have the correct number of samples and loci.

            The sequence data must have the same length for each sample.
        """
        kwargs = {"prefix": prefix, "verbose": self.verbose, "debug": debug}
        logman = LoggerManager(name=__name__, **kwargs)
        self.logger = logman.get_logger()

        # Initialize the parent class GenotypeData
        super().__init__(
            filename=filename,
            filetype="phylip",
            popmapfile=popmapfile,
            force_popmap=force_popmap,
            exclude_pops=exclude_pops,
            include_pops=include_pops,
            guidetree=guidetree,
            qmatrix_iqtree=qmatrix_iqtree,
            qmatrix=qmatrix,
            siterates=siterates,
            siterates_iqtree=siterates_iqtree,
            plot_format=plot_format,
            prefix=prefix,
            verbose=verbose,
            logger=self.logger,
            debug=debug,
        )

    def load_aln(self) -> None:
        """Load the PHYLIP file and populate SNP data, samples, and alleles.

        This method reads the PHYLIP file and populates the SNP data, samples, and alleles. The PHYLIP format is a simple text format for representing multiple sequence alignments. The first line of a PHYLIP file contains the number of samples and the number of loci. Each subsequent line contains the sample ID followed by the sequence data. The sequence data can be in any format, but it is typically a string of nucleotides or amino acids.

        Raises:
            AlignmentFileNotFoundError: If the PHYLIP file is not found.
            AlignmentFormatError: If the PHYLIP file has an invalid format.
        """
        if not self.filename or self.filename is None:
            msg = "No filename provided for PHYLIP file."
            self.logger.error(msg)
            raise AlignmentFormatError(msg)

        if not Path(self.filename).is_file():
            raise AlignmentFileNotFoundError(self.filename)

        self.logger.info(f"Reading PHYLIP file {self.filename}...")

        snp_data = []
        try:
            with open(self.filename, "r") as fin:
                header = fin.readline().strip()
                if not header:
                    msg = "PHYLIP file header is missing or empty."
                    self.logger.error(msg)
                    raise AlignmentFormatError(msg)

                try:
                    n_samples, n_loci = map(int, header.split())
                except ValueError:
                    msg = "Invalid PHYLIP header provided."
                    self.logger.error(msg)
                    raise AlignmentFormatError(msg)

                for line in fin:
                    line = line.strip()
                    if not line:  # Skip blank lines
                        continue
                    cols = line.split()
                    if len(cols) < 2:
                        msg = f"PHYLIP file line does not contain enough columns. Expected 2, but got: {len(cols)}"
                        self.logger.error(msg)
                        raise AlignmentFormatError(msg)

                    if len(cols) > 2:
                        msg = f"PHYLIP file contains too many columns. Expected 2, but got: {len(cols)}"
                        self.logger.error(msg)
                        raise AlignmentFormatError(msg)

                    inds, seqs = cols[0], cols[1]
                    if len(seqs) != n_loci:
                        raise SequenceLengthError(inds)

                    snp_data.append(list(seqs))
                    if inds not in self.samples:
                        self.samples.append(inds)

            self.snp_data = snp_data

            if (
                n_samples != len(snp_data)
                or len(self.samples) != len(snp_data)
                or len(self.samples) != n_samples
            ):
                msg = "Unexpected number of samples encountered."
                self.logger.error(msg)
                PhylipAlignmentSampleMismatch(
                    n_samples, len(self.samples), len(snp_data)
                )
            self._ref, self._alt, self._alt2 = self._get_ref_alt_alleles(
                np.array(snp_data)
            )

            self.logger.info(f"PHYLIP file successfully loaded!")
            self.logger.info(
                f"Found {self.num_snps} SNPs and {self.num_inds} individuals."
            )
        except (AlignmentError, Exception) as e:
            msg = f"An error occurred while reading the PHYLIP file: {e}"
            self.logger.error(msg)
            raise e

        self._validate_seq_lengths()

    def write_phylip(
        self,
        output_file: str,
        genotype_data: Any = None,
        snp_data: Optional[List[List[str]]] = None,
        samples: Optional[List[str]] = None,
    ) -> None:
        """Write the stored alignment as a PHYLIP file.

        This method writes the stored alignment as a PHYLIP file. The PHYLIP format is a simple text format for representing multiple sequence alignments. The first line of a PHYLIP file contains the number of samples and the number of loci. Each subsequent line contains the sample ID followed by the sequence data.

        Args:
            output_file (str): Name of the output PHYLIP file.
            genotype_data (GenotypeData, optional): GenotypeData instance.
            snp_data (List[List[str]], optional): SNP data. Must be provided if genotype_data is None.
            samples (List[str], optional): List of sample IDs. Must be provided if snp_data is not None.

        Raises:
            TypeError: If genotype_data and snp_data are both provided.
            TypeError: If samples are not provided when snp_data is provided.
            ValueError: If samples and snp_data are not the same length.

        Note:
            If genotype_data is provided, the snp_data and samples are loaded from the GenotypeData instance.

            If snp_data is provided, the samples must also be provided.

            If genotype_data is not provided, the snp_data and samples must be provided.

            The sequence data must have the same length for each sample.

            The PHYLIP file must have the correct number of samples and loci.
        """
        self.logger.info(f"Writing to PHYLIP file {output_file}...")

        if genotype_data is not None and snp_data is not None:
            msg = "'genotype_data' and 'snp_data' cannot both be provided."
            self.logger.error(msg)
            raise TypeError(msg)
        elif genotype_data is None and snp_data is None:
            snp_data = self.snp_data
            samples = self.samples
        elif genotype_data is not None and snp_data is None:
            snp_data = genotype_data.snp_data
            samples = genotype_data.samples
        elif genotype_data is None and snp_data is not None:
            if samples is None:
                msg = "'samples' must be provided if 'snp_data' is provided."
                self.logger.error(msg)
                raise TypeError(msg)

        self._validate_seq_lengths()

        try:
            with open(output_file, "w") as f:
                n_samples, n_loci = len(samples), len(snp_data[0])
                f.write(f"{n_samples} {n_loci}\n")
                for sample, sample_data in zip(samples, snp_data):
                    genotype_str = "".join(str(x) for x in sample_data)
                    f.write(f"{sample}\t{genotype_str}\n")

            self.logger.info(f"Successfully wrote PHYLIP file {output_file}!")
        except Exception as e:
            msg = f"An error occurred while writing the PHYLIP file: {e}"
            self.logger.error(msg)
            raise

    @property
    def ref(self) -> List[str]:
        """List of reference alleles."""
        return self._ref

    @property
    def alt(self) -> List[str]:
        """List of alternate alleles."""
        return self._alt

    @property
    def alt2(self) -> List[List[str]]:
        """List of second alternate alleles."""
        return self._alt2



from pathlib import Path
from typing import List, Optional

import numpy as np

from snpio.read_input.genotype_data import GenotypeData
from snpio.utils.custom_exceptions import (
    AlignmentError,
    AlignmentFileNotFoundError,
    AlignmentFormatError,
    StructureAlignmentSampleMismatch,
)
from snpio.utils.logging import LoggerManager


class StructureReader(GenotypeData):
    """This class reads STRUCTURE files and stores the SNP data, sample IDs, and populations.

    ``StructureReader`` is a subclass of GenotypeData and inherits its attributes and methods. It reads STRUCTURE files and stores the SNP data, sample IDs, and populations, as well as various other attributes.

    Example:
        >>> from snpio import StructureReader
        >>>
        >>> genotype_data = StructureReader(filename="data.structure", popmapfile="example.popmap", verbose=True)
        >>>
        >>> genotype_data.snp_data
        array([["A", "T", "T", "A"], ["C", "G", "G", "C"], ["A", "T", "T", "A"]], dtype="<U1")
        >>>
        >>> genotype_data.samples
        ["Sample1", "Sample2", "Sample3", "Sample4"]
        >>>
        >>> genotype_data.populations
        ["Pop1", "Pop1", "Pop2", "Pop2"]
        >>>
        >>> genotype_data.num_snps
        3
        >>>
        >>> genotype_data.num_inds
        4
        >>>
        >>> genotype_data.popmap
        >>> {"Sample1": "Pop1", "Sample2": "Pop1", "Sample3": "Pop2", "Sample4": "Pop2"}
        >>>
        >>> genotype_data.popmap_inverse
        {"Pop1": ["Sample1", "Sample2"], "Pop2": ["Sample3", "Sample4"]}
        >>>
        >>> genotype_data.ref
        ["A", "C", "A"]
        >>>
        >>> genotype_data.alt
        ["T", "G", "T"]
        >>>
        >>> genotype_data.missingness_reports()
        >>>
        >>> genotype_data.run_pca()
        >>>
        >>> genotype_data.write_structure("output.str")

    Attributes:
        logger (LoggerManager): Logger object.
        _has_popids (bool): If True, the STRUCTURE file includes population IDs.
        _onerow (bool): If True, the STRUCTURE file is in one-row format.
    """

    def __init__(
        self,
        filename: Optional[str] = None,
        popmapfile: Optional[str] = None,
        has_popids: bool = False,
        force_popmap: bool = False,
        exclude_pops: Optional[List[str]] = None,
        include_pops: Optional[List[str]] = None,
        guidetree: Optional[str] = None,
        qmatrix_iqtree: Optional[str] = None,
        qmatrix: Optional[str] = None,
        siterates: Optional[str] = None,
        siterates_iqtree: Optional[str] = None,
        plot_format: Optional[str] = "png",
        prefix="snpio",
        verbose: bool = False,
        debug: bool = False,
    ) -> None:
        """Initialize the StructureReader class.

        This class reads STRUCTURE files and stores the SNP data, sample IDs, and populations. It is a subclass of GenotypeData and inherits its attributes and methods.

        Args:
            filename (str, optional): Name of the STRUCTURE file.
            popmapfile (str, optional): Name of the population map file.
            has_popids (bool, optional): If True, the STRUCTURE file includes population IDs.
            force_popmap (bool, optional): If True, force the use of the popmap file.
            exclude_pops (List[str], optional): List of populations to exclude.
            include_pops (List[str], optional): List of populations to include.
            guidetree (str, optional): Name of the guide tree file.
            qmatrix_iqtree (str, optional): Name of the Q-matrix file generated by IQ-TREE.
            qmatrix (str, optional): Name of the Q-matrix file.
            siterates (str, optional): Name of the site rates file.
            siterates_iqtree (str, optional): Name of the site rates file generated by IQ-TREE.
            plot_format (str, optional): Format for saving plots (e.g., 'png', 'svg').
            prefix (str, optional): Prefix for output files.
            verbose (bool, optional): If True, status updates are printed.
            debug (bool, optional): If True, debug messages are printed.

        Raises:
            AlignmentFileNotFoundError: If the STRUCTURE file is not found.

        Note:
            If the filename is provided, the STRUCTURE file is loaded immediately.

            If the filename is not provided, the STRUCTURE file can be loaded later using the load_aln method.

            The STRUCTURE file can be written to a new file using the write_structure method.
        """

        # Set up logger
        kwargs = dict(prefix=prefix, verbose=verbose, debug=debug)
        logman = LoggerManager(name=__name__, **kwargs)
        self.logger = logman.get_logger()

        # Initialize the parent class GenotypeData
        super().__init__(
            filename=filename,
            filetype="structure",
            popmapfile=popmapfile,
            force_popmap=force_popmap,
            exclude_pops=exclude_pops,
            include_pops=include_pops,
            guidetree=guidetree,
            qmatrix_iqtree=qmatrix_iqtree,
            qmatrix=qmatrix,
            siterates=siterates,
            siterates_iqtree=siterates_iqtree,
            plot_format=plot_format,
            prefix=prefix,
            verbose=verbose,
            logger=self.logger,
            debug=debug,
        )

        self._has_popids: bool = has_popids
        self._onerow: bool = False

    def load_aln(self) -> None:
        """Load the STRUCTURE file and populate SNP data, samples, and populations.

        This method reads the STRUCTURE file and populates the SNP data, sample IDs, and populations. It also sets the number of SNPs and individuals.

        Raises:
            AlignmentNotFoundError: If the STRUCTURE file is not found.
            AlignmentFormatError: If the STRUCTURE file has an invalid format.

        Note:
            This method should be called after initializing the StructureReader object.

            The SNP data, sample IDs, and populations are stored in the snp_data, samples, and populations attributes, respectively.4

            The number of SNPs and individuals are stored in the num_snps and num_inds attributes, respectively.

            The STRUCTURE file can be written to a new file using the write_structure method.

            The SNP data can be accessed in the IUPAC format using the `snp_data` attribute.
        """
        if not self.filename:
            raise AlignmentFileNotFoundError(self.filename)

        if not Path(self.filename).is_file():
            raise AlignmentFileNotFoundError(self.filename)

        self.logger.info(f"Reading STRUCTURE file {self.filename}...")

        samples = []
        snp_data = []

        try:
            self._onerow = self._detect_format()

            with open(self.filename, "r") as fin:
                lines = fin.readlines()

            if self._onerow:
                for line in lines:
                    line = line.strip().split()
                    samples.append(line[0])
                    alleles = line[1:] if not self._has_popids else line[2:]
                    snp_data.append(alleles)
            else:
                for i in range(0, len(lines), 2):
                    firstline = lines[i].strip().split()
                    secondline = lines[i + 1].strip().split()

                    if firstline[0] != secondline[0]:
                        raise AlignmentFormatError(
                            f"Sample names do not match between lines: "
                            f"{firstline[0]} and {secondline[0]}"
                        )

                    samples.append(firstline[0])
                    alleles1 = firstline[1:] if not self._has_popids else firstline[2:]
                    alleles2 = (
                        secondline[1:] if not self._has_popids else secondline[2:]
                    )

                    merged_alleles = np.array(
                        [f"{a1}/{a2}" for a1, a2 in zip(alleles1, alleles2)]
                    )
                    snp_data.append(merged_alleles)

            self.samples = np.unique(samples).tolist()

            if len(self.samples) != len(snp_data):
                self.logger.error("Unexpected number of samples found.")
                raise StructureAlignmentSampleMismatch(len(self.samples), len(snp_data))

            self.snp_data = [
                list(map(self._genotype_to_iupac, row)) for row in snp_data
            ]

            self.logger.info(f"STRUCTURE file successfully loaded!")
            self.logger.info(
                f"Found {self.num_snps} SNPs and {self.num_inds} individuals."
            )

        except (AlignmentError, Exception) as e:
            msg = f"An error occurred while reading the STRUCTURE file: {e}"
            self.logger.error(msg)
            raise

    def write_structure(
        self,
        output_file: str,
        genotype_data=None,
        snp_data: Optional[List[List[str]]] = None,
        samples: Optional[List[str]] = None,
        verbose: bool = False,
    ) -> None:
        """Write the stored alignment as a STRUCTURE file.

        This method writes the stored alignment as a STRUCTURE file. If genotype_data is provided, the SNP data and sample IDs are extracted from it. Otherwise, the SNP data and sample IDs must be provided.

        Args:
            output_file (str): Name of the output STRUCTURE file.
            genotype_data (GenotypeData, optional): GenotypeData instance.
            snp_data (List[List[str]], optional): SNP data in IUPAC format. Must be provided if genotype_data is None.
            samples (List[str]], optional): List of sample IDs. Must be provided if snp_data is not provided.
            verbose (bool, optional): If True, status updates are printed.

        Raises:
            TypeError: If genotype_data and snp_data are both provided.
            TypeError: If samples are not provided when snp_data is provided.
        """
        self.logger.info(f"\nWriting STRUCTURE file {output_file}...")

        if genotype_data is not None and snp_data is None:
            snp_data = genotype_data.snp_data
            samples = genotype_data.samples
        elif genotype_data is None and snp_data is None:
            snp_data = self.snp_data
            samples = self.samples
        elif genotype_data is None and snp_data is not None:
            if samples is None:
                raise TypeError("samples must be provided if snp_data is provided.")

        try:
            with open(output_file, "w") as fout:
                for sample, sample_data in zip(samples, snp_data):
                    # Convert IUPAC codes back to genotype format (e.g., 1/1, 2/2)
                    genotypes = [
                        self._iupac_to_genotype(iupac) for iupac in sample_data
                    ]

                    if self._onerow:
                        # Flatten the genotype pairs and write to file in one-row format
                        genotype_pairs = [
                            allele
                            for genotype in genotypes
                            for allele in genotype.split("/")
                        ]
                        fout.write(f"{sample}\t" + "\t".join(genotype_pairs) + "\n")
                    else:
                        # Write the two alleles in two separate rows for two-row format
                        first_row = [genotype.split("/")[0] for genotype in genotypes]
                        second_row = [genotype.split("/")[1] for genotype in genotypes]
                        fout.write(f"{sample}\t" + "\t".join(first_row) + "\n")
                        fout.write(f"{sample}\t" + "\t".join(second_row) + "\n")

            self.logger.info("Successfully wrote STRUCTURE file!")
        except Exception as e:
            msg = f"An error occurred while writing the STRUCTURE file: {e}"
            self.logger.error(msg)
            raise

    def _detect_format(self) -> bool:
        """Detect the format of the STRUCTURE file (onerow or tworow).

        This method reads the first two lines of the STRUCTURE file and compares the sample names. If the sample names are the same, the file is in one-row format; otherwise, it is in two-row format.

        Returns:
            bool: True if the file is in one-row format, False if it is in two-row format.
        """
        if not Path(self.filename).is_file():
            msg = f"STRUCTURE file {self.filename} not found."
            self.logger.error(msg)
            raise AlignmentFileNotFoundError(self.filename)

        with open(self.filename, "r") as fin:
            first_line = fin.readline().split()
            second_line = fin.readline().split()

        # If the first two lines have the same sample name, then it's a two-row format
        onerow = first_line[0] != second_line[0]
        return onerow



import re
import tempfile
import textwrap
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import h5py
import numpy as np
import pysam
from tqdm import tqdm

from snpio.read_input.genotype_data import GenotypeData
from snpio.utils.benchmarking import Benchmark
from snpio.utils.logging import LoggerManager

measure_execution_time = Benchmark.measure_execution_time


class VCFReader(GenotypeData):
    """A class to read VCF files into GenotypeData objects and write GenotypeData objects to VCF files.

    This class inherits from GenotypeData and provides methods to read VCF files and extract the necessary attributes.

    Example:
        >>> from snpio import VCFReader
        >>>
        >>> genotype_data = VCFReader(filename="example.vcf", popmapfile="popmap.txt", verbose=True)
        >>> genotype_data.snp_data
        array([["A", "T", "T", "A"], ["A", "T", "T", "A"], ["A", "T", "T", "A"]], dtype="<U1")
        >>>
        >>> genotype_data.samples
        ["sample1", "sample2", "sample3", "sample4"]
        >>>
        >>> genotype_data.num_inds
        4
        >>>
        >>> genotype_data.num_snps
        3
        >>>
        >>> genotype_data.populations
        ["pop1", "pop1", "pop2", "pop2"]
        >>>
        >>> genotype_data.popmap
        {"sample1": "pop1", "sample2": "pop1", "sample3": "pop2", "sample4":
        "pop2"}
        >>>
        >>> genotype_data.popmap_inverse
        {"pop1": ["sample1", "sample2"], "pop2": ["sample3", "sample4"]}
        >>>
        >>> genotype_data.loci_indices
        array([True, True, True], dtype=bool)
        >>>
        >>> genotype_data.sample_indices
        array([True, True, True, True], dtype=bool)
        >>>
        >>> genotype_data.ref
        ["A", "A", "A"]
        >>>
        >>> genotype_data.alt
        ["T", "T", "T"]
        >>>
        >>> genotype_data.missingness_reports()
        >>>
        >>> genotype_data.run_pca()
        >>>
        >>> genotype_data.write_vcf("output.vcf")

    Attributes:
        filename (Optional[str]): The name of the VCF file to read.
        popmapfile (Optional[str]): The name of the population map file to read.
        chunk_size (int): The size of the chunks to read from the VCF file.
        force_popmap (bool): Whether to force the use of the population map file.
        exclude_pops (Optional[List[str]]): The populations to exclude.
        include_pops (Optional[List[str]]): The populations to include.
        plot_format (str): The format to save the plots in.
        plot_fontsize (int): The font size for the plots.
        plot_dpi (int): The DPI for the plots.
        plot_despine (bool): Whether to remove the spines from the plots.
        show_plots (bool): Whether to show the plots.
        prefix (str): The prefix to use for the output files.
        verbose (bool): Whether to print verbose output.
        sample_indices (np.ndarray): The indices of the samples to read.
        loci_indices (np.ndarray): The indices of the loci to read.
        debug (bool): Whether to enable debug mode.
        num_records (int): The number of records in the VCF file.
        filetype (str): The type of the file.
        vcf_header (Optional[pysam.libcbcf.VariantHeader]): The VCF header.
        info_fields (Optional[List[str]]): The VCF info fields.
        resource_data (dict): A dictionary to store resource data.
        logger (logging.Logger): The logger object.
        snp_data (np.ndarray): The SNP data.
        samples (np.ndarray): The sample names.

    Note:
        The VCF file is bgzipped, sorted, and indexed using Tabix to ensure efficient reading, if necessary.

        The VCF file is read in chunks to avoid memory issues.

        The VCF attributes are extracted and stored in an HDF5 file for efficient access.

        The genotype data is transformed into IUPAC codes for efficient storage and processing.

        The VCF attributes are stored in an HDF5 file for efficient access.

        The VCF attributes are extracted and stored in an HDF5 file for efficient access.

        The genotype data is transformed into IUPAC codes for efficient storage and processing.

        The VCF attributes are stored in an HDF5 file for efficient access.
    """

    def __init__(
        self,
        filename: Optional[str] = None,
        popmapfile: Optional[str] = None,
        chunk_size: int = 1000,
        force_popmap: bool = False,
        exclude_pops: Optional[List[str]] = None,
        include_pops: Optional[List[str]] = None,
        plot_format: str = "png",
        plot_fontsize: int = 18,
        plot_dpi: int = 300,
        plot_despine: bool = True,
        show_plots: bool = False,
        prefix: str = "snpio",
        verbose: bool = False,
        sample_indices: Optional[np.ndarray] = None,
        loci_indices: Optional[np.ndarray] = None,
        debug: bool = False,
    ) -> None:
        """Initializes the VCFReader object.

        This class inherits from GenotypeData and provides methods to read VCF files and extract the necessary attributes.

        Args:
            filename (Optional[str], optional): The name of the VCF file to read. Defaults to None.
            popmapfile (Optional[str], optional): The name of the population map file to read. Defaults to None.
            chunk_size (int, optional): The size of the chunks to read from the VCF file. Defaults to 1000.
            force_popmap (bool, optional): Whether to force the use of the population map file. Defaults to False.
            exclude_pops (Optional[List[str]], optional): The populations to exclude. Defaults to None.
            include_pops (Optional[List[str]], optional): The populations to include. Defaults to None.
            plot_format (str, optional): The format to save the plots in. Defaults to "png".
            plot_fontsize (int, optional): The font size for the plots. Defaults to 18.
            plot_dpi (int, optional): The DPI for the plots. Defaults to 300.
            plot_despine (bool, optional): Whether to remove the spines from the plots. Defaults to True.
            show_plots (bool, optional): Whether to show the plots. Defaults to False.
            prefix (str, optional): The prefix to use for the output files. Defaults to "snpio".
            verbose (bool, optional): Whether to print verbose output. Defaults to False.
            sample_indices (Optional[np.ndarray], optional): The indices of the samples to read. Defaults to None.
            loci_indices (Optional[np.ndarray], optional): The indices of the loci to read. Defaults to None.
            debug (bool, optional): Whether to enable debug mode. Defaults to False.
        """
        outdir = Path(f"{prefix}_output") / "gtdata" / "alignments" / "vcf"
        self._vcf_attributes_fn = outdir / "vcf_attributes.h5"

        kwargs = {"prefix": prefix, "verbose": verbose, "debug": debug}
        logman = LoggerManager(name=__name__, **kwargs)
        self.logger = logman.get_logger()

        self.resource_data = {}

        super().__init__(
            filename=filename,
            filetype="vcf",
            popmapfile=popmapfile,
            force_popmap=force_popmap,
            exclude_pops=exclude_pops,
            include_pops=include_pops,
            plot_format=plot_format,
            plot_fontsize=plot_fontsize,
            plot_dpi=plot_dpi,
            plot_despine=plot_despine,
            show_plots=show_plots,
            prefix=prefix,
            verbose=verbose,
            sample_indices=sample_indices,
            loci_indices=loci_indices,
            chunk_size=chunk_size,
            logger=self.logger,
            debug=debug,
        )

        self.num_records = 0
        self.filetype = "vcf"
        self.vcf_header = None
        self.info_fields = None

        outdir.mkdir(exist_ok=True, parents=True)

        # Define reverse IUPAC mapping (example, should be defined appropriately)
        self.reverse_iupac_mapping = {
            "A": ("A", "A"),
            "C": ("C", "C"),
            "G": ("G", "G"),
            "T": ("T", "T"),
            "M": ("A", "C"),
            "R": ("A", "G"),
            "W": ("A", "T"),
            "S": ("C", "G"),
            "Y": ("C", "T"),
            "K": ("G", "T"),
            "V": ("A", "C", "G"),
            "H": ("A", "C", "T"),
            "D": ("A", "G", "T"),
            "B": ("C", "G", "T"),
            "N": ("A", "C", "G", "T"),
        }

    @measure_execution_time
    def load_aln(self) -> None:
        """Loads the alignment from the VCF file into the VCFReader object.

        This method ensures that the input VCF file is bgzipped, sorted, and indexed using Tabix. It then reads the VCF file and extracts the necessary attributes. The VCF attributes are stored in an HDF5 file for efficient access. The genotype data is transformed into IUPAC codes for efficient storage and processing.
        """
        self.logger.info(f"Processing input VCF file {self.filename}...")

        vcf_path = Path(self.filename)
        bgzipped_vcf = vcf_path

        # Ensure the VCF file is bgzipped
        if not self._is_bgzipped(vcf_path):
            bgzipped_vcf = self._bgzip_file(vcf_path)
            self.logger.info(f"Bgzipped VCF file to {bgzipped_vcf}")

        # Ensure the VCF file is sorted
        if not self._is_sorted(bgzipped_vcf):
            bgzipped_vcf = self._sort_vcf_file(bgzipped_vcf)
            self.logger.info(f"Sorted VCF file to {bgzipped_vcf}")

        # Ensure the VCF file is indexed
        if not self._has_tabix_index(bgzipped_vcf):
            self._tabix_index(bgzipped_vcf)
            self.logger.info(f"Indexed VCF file {bgzipped_vcf}.tbi")

        self.filename = str(bgzipped_vcf)

        with pysam.VariantFile(self.filename, mode="r") as vcf:
            self.vcf_header = vcf.header
            self.samples = np.array(vcf.header.samples)
            self.num_records = sum(1 for _ in vcf)
            vcf.reset()

            self.vcf_attributes_fn, self.snp_data, self.samples = (
                self.get_vcf_attributes(vcf, chunk_size=self.chunk_size)
            )

            self.snp_data = self.snp_data.T

        self.logger.info(f"VCF file successfully loaded!")

        msg = (
            f"\nLoaded Alignment contains:\n"
            f"{self.snp_data.shape[0]} samples,\n"
            f"{self.snp_data.shape[1]} loci.\n"
        )
        self.logger.info(msg)

        self.logger.debug(f"snp_data: {self.snp_data}")
        self.logger.debug(f"snp_data shape: {self.snp_data.shape}")

        return self

    def get_vcf_attributes(
        self, vcf: pysam.VariantFile, chunk_size: int = 1000
    ) -> Tuple[str, np.ndarray, np.ndarray]:
        """Extracts VCF attributes and returns them in an efficient manner with chunked processing.

        Args:
            vcf (pysam.VariantFile): The VCF file object to extract attributes from.
            chunk_size (int, optional): The size of the chunks to process. Defaults to 1000.

        Returns:
            Tuple[str, np.ndarray, np.ndarray]: The path to the HDF5 file containing the VCF attributes, the SNP data, and the sample names.
        """
        h5_outfile = self.vcf_attributes_fn
        h5_outfile.parent.mkdir(exist_ok=True, parents=True)

        snp_data = []
        sample_names = np.array(vcf.header.samples)

        with h5py.File(h5_outfile, "w") as f:
            chrom_dset = f.create_dataset(
                "chrom", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )
            pos_dset = f.create_dataset("pos", (0,), maxshape=(None,), dtype=np.int32)
            id_dset = f.create_dataset(
                "id", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )
            ref_dset = f.create_dataset("ref", (0,), maxshape=(None,), dtype="S1")
            alt_dset = f.create_dataset(
                "alt", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )

            qual_dset = f.create_dataset(
                "qual", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )

            filt_dset = f.create_dataset("filt", (0,), maxshape=(None,), dtype="S4")

            fmt_dset = f.create_dataset(
                "fmt", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )

            fmt_data_dset = f.create_dataset(
                "fmt_data", (0,), maxshape=(None,), dtype=h5py.string_dtype()
            )

            info_fields = list(vcf.header.info)
            self.info_fields = info_fields

            info_group = f.create_group("info")

            # Create datasets within info groups
            info_dsets = {
                k: info_group.create_dataset(
                    k, (0,), maxshape=(None,), dtype=h5py.string_dtype()
                )
                for k in info_fields
            }

            (
                chrom_chunk,
                pos_chunk,
                id_chunk,
                ref_chunk,
                alt_chunk,
                qual_chunk,
                filt_chunk,
                fmt_chunk,
                info_chunk,
                snp_chunk,
                fmt_data_chunk,
            ) = ([], [], [], [], [], [], [], [], defaultdict(list), [], [])

            for variant in tqdm(
                vcf.fetch(),
                desc="Processing VCF records: ",
                total=self.num_records,
                unit=" records",
            ):
                chrom_chunk.append(variant.chrom)
                pos_chunk.append(variant.pos)
                id_chunk.append("." if variant.id is None else variant.id)
                ref_chunk.append(variant.ref)

                if variant.alts is None:
                    alt_chunk.append(".")
                else:
                    alt_chunk.append(",".join([a for a in variant.alts]))

                qual_chunk.append("." if variant.qual is None else str(variant.qual))
                filt_chunk.append(
                    "." if not variant.filter.keys() else list(variant.filter.keys())[0]
                )
                fmt_chunk.append(":".join([k for k in variant.format.keys()]))

                for k in info_fields:
                    value = variant.info.get(k, ".")
                    processed_value = (
                        ",".join(list(value)) if isinstance(value, tuple) else value
                    )
                    info_chunk[k].append(processed_value)

                gt_data = [
                    variant.samples[sample].get("GT", (None, None))
                    for sample in sample_names
                ]

                fmt_data = []
                for sample in sample_names:
                    fmd_list = []
                    for k in variant.format.keys():
                        if k == "GT":
                            continue
                        fmd = variant.samples[sample].get(k, ".")
                        if isinstance(fmd, (tuple, list)):
                            if isinstance(fmd, tuple):
                                fmd = list(fmd)
                            fmd = ",".join([str(x) for x in fmd])
                        else:
                            fmd = str(fmd)
                        fmd_list.append(fmd)
                    fmt_data.append(":".join(fmd_list))
                fmt_data_chunk.append("\t".join(fmt_data))

                transformed_gt = self.transform_gt(
                    np.array(gt_data), variant.ref, variant.alts
                )
                snp_chunk.append(transformed_gt)

                # Process chunk if it reaches the specified chunk size
                if len(chrom_chunk) == chunk_size:
                    # Extend the datasets
                    current_size = chrom_dset.shape[0]
                    new_size = current_size + chunk_size
                    chrom_dset.resize((new_size,))
                    pos_dset.resize((new_size,))
                    id_dset.resize((new_size,))
                    ref_dset.resize((new_size,))
                    alt_dset.resize((new_size,))
                    qual_dset.resize((new_size,))
                    filt_dset.resize((new_size,))
                    fmt_dset.resize((new_size,))
                    fmt_data_dset.resize((new_size,))

                    for k in info_dsets.keys():
                        if info_chunk[k]:
                            info_dsets[k].resize((new_size,))
                            info_dsets[k][current_size:new_size] = np.array(
                                info_chunk[k], dtype=str
                            )
                            info_chunk[k] = []

                    # Write the chunk data
                    chrom_dset[current_size:new_size] = chrom_chunk
                    pos_dset[current_size:new_size] = pos_chunk
                    id_dset[current_size:new_size] = id_chunk
                    ref_dset[current_size:new_size] = ref_chunk
                    alt_dset[current_size:new_size] = alt_chunk

                    qual_dset[current_size:new_size] = qual_chunk
                    filt_dset[current_size:new_size] = filt_chunk
                    fmt_dset[current_size:new_size] = fmt_chunk
                    fmt_data_dset[current_size:new_size] = fmt_data_chunk

                    snp_data.extend(snp_chunk)

                    # Reset the chunk lists
                    (
                        chrom_chunk,
                        pos_chunk,
                        id_chunk,
                        ref_chunk,
                        alt_chunk,
                        qual_chunk,
                        filt_chunk,
                        fmt_chunk,
                        snp_chunk,
                        fmt_data_chunk,
                    ) = (
                        [],
                        [],
                        [],
                        [],
                        [],
                        [],
                        [],
                        [],
                        [],
                        [],
                    )

            # Handle any remaining data in the chunks
            if chrom_chunk:
                current_size = chrom_dset.shape[0]
                new_size = current_size + len(chrom_chunk)
                chrom_dset.resize((new_size,))
                pos_dset.resize((new_size,))
                id_dset.resize((new_size,))
                ref_dset.resize((new_size,))
                alt_dset.resize((new_size,))
                qual_dset.resize((new_size,))
                filt_dset.resize((new_size,))
                fmt_dset.resize((new_size,))
                fmt_data_dset.resize((new_size,))

                for k in info_dsets.keys():
                    info_dsets[k].resize((new_size,))
                    info_dsets[k][current_size:new_size] = np.array(
                        info_chunk[k], dtype=str
                    )
                    info_chunk[k] = []

                chrom_dset[current_size:new_size] = chrom_chunk
                pos_dset[current_size:new_size] = pos_chunk
                id_dset[current_size:new_size] = id_chunk
                ref_dset[current_size:new_size] = ref_chunk
                alt_dset[current_size:new_size] = alt_chunk
                qual_dset[current_size:new_size] = qual_chunk
                filt_dset[current_size:new_size] = filt_chunk
                fmt_dset[current_size:new_size] = fmt_chunk
                fmt_data_dset[current_size:new_size] = fmt_data_chunk

                snp_data.extend(snp_chunk)

        return h5_outfile, np.array(snp_data), sample_names

    def transform_gt(self, gt: np.ndarray, ref: str, alts: List[str]) -> np.ndarray:
        """Transforms genotype tuples into their IUPAC codes or corresponding strings.

        Args:
            gt (np.ndarray): The genotype tuples to transform.
            ref (str): The reference allele.
            alts (List[str]): The alternate alleles.

        Returns:
            np.ndarray: The transformed genotype tuples.
        """
        iupac_mapping = self._iupac_code()

        # Convert alts to a list, ensuring it's mutable and indexable
        alts = list(alts) if alts else []

        # Prepare the array of possible alleles including reference
        alt_array = np.array([ref] + alts)

        transformed_gt = np.array(
            [
                (
                    iupac_mapping(
                        (
                            (
                                alt_array[allele1]
                                if 0 <= allele1 < len(alt_array)
                                else ref
                            ),
                            (
                                alt_array[allele2]
                                if 0 <= allele2 < len(alt_array)
                                else ref
                            ),
                        )
                    )
                    if allele1 is not None and allele2 is not None
                    else "N"
                )
                for allele1, allele2 in gt
            ]
        )
        return transformed_gt

    def _iupac_code(self) -> Callable:
        """Returns a callable function to get the IUPAC code for a given allele pair.

        Returns:
            Callable: A function to get the IUPAC code for a given allele pair.
        """
        iupac_mapping = self._iupac_from_gt_tuples()

        def get_code(alleles: Tuple[str]) -> str:
            return iupac_mapping.get(tuple(sorted(alleles)), "N")

        return get_code

    def _iupac_from_gt_tuples(self) -> Dict[Tuple[str], str]:
        """Creates a mapping from genotype tuples to IUPAC codes.

        Returns:
            Dict[Tuple[str], str]: A dictionary mapping genotype tuples to IUPAC codes.
        """
        # Define IUPAC codes for allele pairs
        mapping = {
            ("A", "A"): "A",
            ("A", "C"): "M",
            ("A", "G"): "R",
            ("A", "T"): "W",
            ("C", "C"): "C",
            ("C", "G"): "S",
            ("C", "T"): "Y",
            ("G", "G"): "G",
            ("G", "T"): "K",
            ("T", "T"): "T",
        }
        return mapping

    def _is_bgzipped(self, filepath: Path) -> bool:
        """Checks if a file is bgzipped.

        Args:
            filepath (Path): The path to the file.

        Returns:
            bool: True if bgzipped, False otherwise.
        """
        try:
            with open(filepath, "rb") as f:
                magic = f.read(2)
            return magic == b"\x1f\x8b"
        except Exception as e:
            self.logger.error(f"Error checking bgzip status: {e}")
            return False

    def _bgzip_file(self, filepath: Path) -> Path:
        """BGZips a VCF file using pysam's BGZFile.

        Args:
            filepath (Path): The path to the VCF file.

        Returns:
            Path: The path to the bgzipped VCF file.
        """
        bgzipped_path = filepath.with_suffix(filepath.suffix + ".gz")
        try:
            with open(filepath, "rb") as f_in, pysam.BGZFile(
                str(bgzipped_path), "wb"
            ) as f_out:
                # Read and write in chunks to handle large files efficiently
                chunk_size = 1024 * 1024  # 1MB
                while True:
                    chunk = f_in.read(chunk_size)
                    if not chunk:
                        break
                    f_out.write(chunk)
            return bgzipped_path
        except Exception as e:
            self.logger.error(f"Error bgzipping file {filepath}: {e}")
            raise

    def _is_sorted(self, filepath: Path) -> bool:
        """Checks if the VCF file is sorted alphanumerically by chromosome and position.

        Args:
            filepath (Path): The path to the bgzipped VCF file.

        Returns:
            bool: True if sorted, False otherwise.
        """
        try:
            with pysam.VariantFile(filepath, mode="r") as vcf:
                previous_chrom = None
                previous_pos = -1
                for record in vcf:
                    if previous_chrom is not None:
                        if record.chrom < previous_chrom:
                            return False
                        elif (
                            record.chrom == previous_chrom and record.pos < previous_pos
                        ):
                            return False
                    previous_chrom = record.chrom
                    previous_pos = record.pos
            return True
        except Exception as e:
            self.logger.error(f"Error checking sort order: {e}")
            return False

    @staticmethod
    def _natural_sort_key(chrom_pos: Tuple[str, int]) -> Any:
        """Extracts a natural sort key for sorting."""
        chrom, pos = chrom_pos
        # Split chrom into alphanumeric segments to achieve natural order
        return [int(s) if s.isdigit() else s for s in re.split("([0-9]+)", chrom)], int(
            pos
        )

    def _sort_vcf_file(self, filepath: Path) -> Path:
        """Sorts a VCF file alphanumerically using custom logic, then indexes it.

        Args:
            filepath (Path): The path to the bgzipped VCF file.

        Returns:
            Path: The path to the sorted bgzipped VCF file.
        """
        sorted_path = filepath.with_name(filepath.stem + "_sorted.vcf.gz")

        try:
            # Read the VCF file and split into header and data lines
            header_lines = []
            data_lines = []
            with pysam.VariantFile(filepath, "r") as vcf_in:
                for line in vcf_in.header.records:
                    header_lines.append(str(line))
                for record in vcf_in:
                    data_lines.append(record)

            # Sort the data lines alphanumerically by CHROM and POS
            sorted_data_lines = sorted(
                data_lines, key=lambda r: VCFReader._natural_sort_key((r.contig, r.pos))
            )

            # Write sorted data to a temporary VCF file
            with tempfile.NamedTemporaryFile(
                delete=False, mode="w", suffix=".vcf"
            ) as temp_vcf:
                # Write the header lines
                for line in header_lines:
                    temp_vcf.write(line + "\n")

                # Write sorted data lines
                for record in sorted_data_lines:
                    temp_vcf.write(str(record) + "\n")

                temp_vcf_path = Path(temp_vcf.name)

            # Bgzip and tabix index the sorted VCF file
            sorted_bgzip_path = sorted_path
            pysam.bgzip(str(temp_vcf_path), str(sorted_bgzip_path))
            pysam.tabix_index(str(sorted_bgzip_path), preset="vcf", force=True)

            # Clean up the temporary uncompressed VCF
            temp_vcf_path.unlink()

            return sorted_bgzip_path

        except Exception as e:
            self.logger.error(f"Error sorting VCF file with custom sort: {e}")
            raise

    def _has_tabix_index(self, filepath: Path) -> bool:
        """Checks if a Tabix index exists for the given VCF file.

        Args:
            filepath (Path): The path to the bgzipped VCF file.

        Returns:
            bool: True if index exists, False otherwise.
        """
        index_path = filepath.with_suffix(filepath.suffix + ".tbi")
        return index_path.exists()

    def _tabix_index(self, filepath: Path) -> None:
        """Creates a Tabix index for a bgzipped VCF file.

        Args:
            filepath (Path): The path to the bgzipped VCF file.
        """
        try:
            pysam.tabix_index(str(filepath), preset="vcf", force=True)
        except Exception as e:
            self.logger.error(f"Error indexing VCF file {filepath}: {e}")
            raise

    def validate_input_vcf(self, filepath: Path) -> None:
        """Validates the input VCF file to ensure it meets required criteria.

        Args:
            filepath (Path): The path to the bgzipped and sorted VCF file.

        Raises:
            ValueError: If the VCF file does not meet validation criteria.
        """
        try:
            with pysam.VariantFile(filepath, mode="r") as vcf:
                required_fields = {
                    "CHROM",
                    "POS",
                    "ID",
                    "REF",
                    "ALT",
                    "QUAL",
                    "FILTER",
                    "INFO",
                    "FORMAT",
                }
                header_fields = set(vcf.header.records[0].keys())
                if not required_fields.issubset(header_fields):
                    missing = required_fields - header_fields
                    msg = f"VCF file is missing required fields: {missing}"
                    self.logger.error(msg)
                    raise ValueError(msg)

                if "GT" not in vcf.header.formats:
                    msg = "VCF file is missing the GT format field."
                    self.logger.error(msg)
                    raise ValueError(msg)

                if len(vcf.header.samples) == 0:
                    msg = "VCF file contains no samples."
                    self.logger.error(msg)
                    raise ValueError("VCF file contains no samples.")
        except Exception as e:
            self.logger.error(f"Input VCF validation failed: {e}")
            raise

    def validate_output_vcf(self, filepath: Path) -> None:
        """Validates the output VCF file to ensure it was written correctly.

        Args:
            filepath (Path): The path to the bgzipped and indexed output VCF file.

        Raises:
            ValueError: If the VCF file does not meet validation criteria.
        """
        try:
            with pysam.VariantFile(filepath, mode="r") as vcf:
                # Basic checks: header presence, samples match, etc.
                if self.vcf_header is None:
                    msg = "Output VCF file lacks a header."
                    self.logger.error(msg)
                    raise ValueError(msg)

                output_samples = set(vcf.header.samples)
                expected_samples = set(self.samples)
                if not expected_samples.issubset(output_samples):
                    missing = expected_samples - output_samples
                    msg = f"Output VCF is missing samples: {missing}"
                    self.logger.error(msg)
                    raise ValueError(msg)

                # Additional checks can be added as needed
        except Exception as e:
            self.logger.error(f"Output VCF validation failed: {e}")
            raise

    @measure_execution_time
    def write_vcf(
        self,
        output_filename: str,
        hdf5_file_path: Optional[str] = None,
        chunk_size: int = 1000,
    ) -> Any:
        """Writes the GenotypeData object data to a VCF file in chunks.

        This method writes the VCF data, bgzips the output file, indexes it with Tabix,
        and validates the output.

        Args:
            output_filename (str): The name of the output VCF file to write.
            hdf5_file_path (str, optional): The path to the HDF5 file containing the VCF attributes. Defaults to None.
            chunk_size (int, optional): The size of the chunks to read from the HDF5 file. Defaults to 1000.

        Returns:
            VCFReader: The current instance of VCFReader.
        """
        self.logger.info(f"Writing VCF file to: {output_filename}")

        if hdf5_file_path is None:
            hdf5_file_path = self.vcf_attributes_fn

        of = Path(output_filename)

        # Open the HDF5 file for reading and the output VCF file for writing
        with h5py.File(hdf5_file_path, "r") as hdf5_file, open(of, "w") as f:
            # Write the VCF header
            f.write(self._build_vcf_header())

            # Get the total number of loci
            total_loci = len(hdf5_file["chrom"])

            chunk_size = self.chunk_size

            with tqdm(
                total=total_loci, desc="Writing VCF Records: ", unit=" records"
            ) as pbar:
                # Iterate over data in chunks
                for start in range(0, total_loci, chunk_size):
                    end = min(start + chunk_size, total_loci)

                    # Read data in chunks
                    chrom = hdf5_file["chrom"][start:end].astype(str)
                    pos = hdf5_file["pos"][start:end]
                    vid = hdf5_file["id"][start:end].astype(str)
                    ref = hdf5_file["ref"][start:end].astype(str)
                    alt = hdf5_file["alt"][start:end].astype(str)
                    qual = hdf5_file["qual"][start:end].astype(str)
                    filt = hdf5_file["filt"][start:end].astype(str)
                    fmt = hdf5_file["fmt"][start:end].astype(str)
                    fmt_data = hdf5_file["fmt_data"][start:end].astype(str)

                    # Read info fields in chunks
                    info_keys = list(hdf5_file["info"].keys())

                    # Create a dictionary to hold the info fields data
                    info = {
                        k: hdf5_file[f"info/{k}"][start:end].astype(str)
                        for k in info_keys
                    }

                    # Format the info fields into strings
                    info_result = [
                        ";".join([f"{key}={value[i]}" for key, value in info.items()])
                        for i in range(end - start)
                    ]

                    # Write each row to the VCF file
                    for i in range(len(chrom)):
                        # Construct the ALT field by combining alternate alleles
                        alt_alleles_str = alt[i] if alt[i] != "." else "."
                        ref_str = ref[i]
                        alts = alt[i].split(",") if alt[i] != "." else []

                        # Split fmt_data[i] by tab to get individual sample data
                        fmt_data_samples = fmt_data[i].split("\t")

                        # Join fmt_data with snp_data by ":"
                        combined_data = [
                            f"{self.snp_data[j, start + i]}:{fmt_data_samples[j]}"
                            for j in range(len(self.samples))
                        ]

                        row = [
                            chrom[i],
                            str(pos[i]),
                            vid[i],
                            ref_str,
                            alt_alleles_str,
                            qual[i],
                            filt[i],
                            info_result[i],
                            fmt[i],
                        ] + combined_data

                        # Replace alleles if necessary
                        row = self._replace_alleles(row, ref_str, alts)

                        # Write the row to the file
                        f.write("\t".join(row) + "\n")

                    pbar.update(end - start)

        # Bgzip and index the output VCF file
        bgzipped_output = self._bgzip_file(of)
        self.logger.info(f"Bgzipped output VCF file to {bgzipped_output}")

        self._tabix_index(bgzipped_output)
        self.logger.info(f"Indexed output VCF file {bgzipped_output}.tbi")
        self.logger.info("Successfully wrote VCF file!")

        return self

    def _replace_alleles(self, row: List[str], ref: str, alts: List[str]) -> List[str]:
        """Replace the alleles in the VCF row with the corresponding VCF genotype codes.

        Args:
            row (List[str]): The VCF row to process.
            ref (str): The reference allele.
            alts (List[str]): The alternate alleles.

        Returns:
            List[str]: The processed VCF row.
        """

        def get_genotype_code(
            allele1: str, allele2: str, ref: str, alts: List[str]
        ) -> str:
            """Assigns the genotype code based on the decoded alleles.

            Args:
                allele1 (str): The first allele.
                allele2 (str): The second allele.
                ref (str): The reference allele.
                alts (List[str]): The alternate alleles.

            Returns:
                str: The genotype code in VCF format.
            """
            if allele1 == ref and allele2 == ref:
                return "0/0"
            elif allele1 == ref and allele2 in alts:
                return f"0/{alts.index(allele2) + 1}"
            elif allele1 in alts and allele2 == ref:
                return f"0/{alts.index(allele1) + 1}"
            elif allele1 in alts and allele2 in alts:
                index1 = alts.index(allele1) + 1
                index2 = alts.index(allele2) + 1
                return f"{min(index1, index2)}/{max(index1, index2)}"
            return "./."

        # Apply allele replacements
        for i in range(9, len(row)):
            data = row[i].split(":")
            iupac_code = data[0]
            fmt_data = data[1:] if len(data) > 1 else []

            # Decode the IUPAC code to get the alleles
            if iupac_code in self.reverse_iupac_mapping:
                alleles = self.reverse_iupac_mapping[iupac_code]
                # If multiple alleles are present in the mapping, take the first two
                if len(alleles) >= 2:
                    allele1, allele2 = alleles[:2]
                else:
                    allele1 = allele2 = alleles[0]
            else:
                # Handle cases where genotype is a single allele or missing
                if iupac_code in {"N", "-", ".", "?"}:
                    row[i] = "./."
                    continue
                else:
                    allele1 = allele2 = iupac_code

            # Assign genotype code based on decoded alleles
            row[i] = get_genotype_code(allele1, allele2, ref, alts)

            if fmt_data:
                fmt_data = ":".join(fmt_data)
                row[i] = f"{row[i]}:{fmt_data}"

        return row

    def _build_vcf_header(self) -> str:
        """Dynamically builds the VCF header based on the loaded sample IDs.

        Returns:
            str: The VCF header.
        """
        sample_headers = "\t".join(self.samples)

        vcf_header = textwrap.dedent(
            f"""\
            ##fileformat=VCFv4.2
            ##source=SNPio
            ##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
            ##INFO=<ID=VAF,Number=A,Type=Float,Description="Variant Allele Frequency">
            ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
            #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{sample_headers}\n"""
        )

        return vcf_header

    def update_vcf_attributes(
        self,
        snp_data: np.ndarray,
        sample_indices: np.ndarray,
        loci_indices: np.ndarray,
        samples: np.ndarray,
    ) -> None:
        """Updates the VCF attributes with new data in chunks.

        Args:
            snp_data (np.ndarray): The SNP data to update the VCF attributes with.
            sample_indices (np.ndarray): The indices of the samples to update.
            loci_indices (np.ndarray): The indices of the loci to update.
            samples (np.ndarray): The sample names to update.

        Raises:
            FileNotFoundError: If the VCF attributes file is not found.
        """
        self.snp_data = snp_data
        self.samples = samples
        self.sample_indices = sample_indices
        self.loci_indices = loci_indices

        hdf5_file_path = Path(self.vcf_attributes_fn)
        hdf5_file_filt = hdf5_file_path.with_name("vcf_attributes_filtered.h5")

        if not hdf5_file_path.exists() and not hdf5_file_path.is_file():
            msg = f"VCF attributes file {hdf5_file_path} not found."
            self.logger.error(msg)
            raise FileNotFoundError(msg)

        new_size = np.count_nonzero(loci_indices)

        # Helper function to process chunks
        def process_chunk(
            dataset_name: str, dtype: Any, fw: h5py.File, fr: h5py.File, new_size: int
        ) -> None:
            """Reads, filters, and writes a dataset in chunks.

            Args:
                dataset_name (str): The name of the dataset to process.
                dtype (object): The datatype of the dataset.
                fw (h5py.File): The filtered output file handle.
                fr (h5py.File): The input file handle.
                new_size (int): The new size of the dataset after filtering.
            """
            chunk_size_inner = self.chunk_size
            dset_size = fr[dataset_name].shape[0]
            fw.create_dataset(
                dataset_name, dtype=dtype, shape=(new_size,), maxshape=(None,)
            )

            # Reading and writing the dataset in chunks
            start_idx = 0
            write_idx = 0
            while start_idx < dset_size:
                end_idx = min(start_idx + chunk_size_inner, dset_size)
                chunk = fr[dataset_name][start_idx:end_idx]
                filtered_chunk = chunk[loci_indices[start_idx:end_idx]]
                chunk_size_to_write = filtered_chunk.size
                fw[dataset_name][
                    write_idx : write_idx + chunk_size_to_write
                ] = filtered_chunk
                write_idx += chunk_size_to_write
                start_idx = end_idx

        # Open the filtered output file for writing
        with h5py.File(hdf5_file_filt, "w") as fw:
            # Process each dataset in chunks
            with h5py.File(hdf5_file_path, "r") as fr:
                pcp = partial(process_chunk, fw=fw, fr=fr, new_size=new_size)
                pcp("chrom", h5py.string_dtype())
                pcp("pos", np.int32)
                pcp("id", h5py.string_dtype())
                pcp("ref", "S1")
                pcp("alt", h5py.string_dtype())
                pcp("qual", h5py.string_dtype())
                pcp("filt", "S4")
                pcp("fmt", h5py.string_dtype())
                pcp("fmt_data", h5py.string_dtype())
                [pcp(f"info/{k}", h5py.string_dtype()) for k in fr["info"].keys()]

        # Update the VCF file path to the new filtered file
        self.vcf_attributes_fn = str(hdf5_file_filt)

    @property
    def vcf_attributes_fn(self) -> str:
        """The path to the HDF5 file containing the VCF attributes.

        Returns:
            str: The path to the HDF5 file containing the VCF attributes.
        """
        return self._vcf_attributes_fn

    @vcf_attributes_fn.setter
    def vcf_attributes_fn(self, value: str) -> None:
        """Sets the path to the HDF5 file containing the VCF attributes.

        Args:
            value (str): The path to the HDF5 file containing the VCF attributes.
        """
        if isinstance(value, str):
            value = Path(value)

        self._vcf_attributes_fn = value



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from snpio.utils import misc, sequence_tools
from snpio.utils.logging import LoggerManager


class GenotypeEncoder:
    """Encode genotypes to various formats suitable for machine learning.

    This class provides methods to encode genotypes to various formats suitable for machine learning, including 012, one-hot, and integer encodings, as well as the inverse operations.

    Example:
        >>> # Import necessary modules
        >>> from snpio import VCFReader, GenotypeEncoder
        >>>
        >>> # Initialize VCFReader and GenotypeEncoder objects
        >>> gd = VCFReader(filename="my_vcf.vcf", popmapfile="my_popmap.txt")
        >>> ge = GenotypeEncoder(gd)
        >>>
        >>> # Encode genotypes to 012, one-hot, and integer formats
        >>> gt_012 = ge.genotypes_012
        >>> gt_onehot = ge.genotypes_onehot(gt_012)
        >>> gt_int = ge.genotypes_int(gt_012)
        >>>
        >>> # Inverse operations
        >>> ge.genotypes_012 = gt_012
        >>> ge.genotypes_onehot = gt_onehot
        >>> ge.genotypes_int = gt_int

    Attributes:
        plot_format (str): Plot format for the data.
        prefix (str): Prefix for the output directory.
        verbose (bool): If True, display verbose output.
        snp_data (List[List[str]]): List of lists of SNPs.
        samples (List[str]): List of sample IDs.
        filetype (str): File type of the data.
        missing_vals (List[str]): List of missing values.
        replace_vals (List[str]): List of values to replace missing values with.
    """

    def __init__(self, genotype_data: Any) -> None:
        """Initialize the GenotypeEncoder object.

        This class provides methods to encode genotypes to various formats suitable for machine learning, including 012, one-hot, and integer encodings, as well as the inverse operations.

        Args:
            genotype_data (GenotypeData): Initialized GenotypeData object.

        Note:
            The GenotypeData object must be initialized before creating an instance of this class.
        """

        self.plot_format = genotype_data.plot_format
        self.prefix = genotype_data.prefix
        self.verbose = genotype_data.verbose
        self.snp_data = genotype_data.snp_data
        self.samples = genotype_data.samples
        debug = genotype_data.debug
        self.filetype = "encoded"

        self.missing_vals = ["N", "-", ".", "?"]
        self.replace_vals = ["-9"] * len(self.missing_vals)

        kwargs = {"verbose": self.verbose, "debug": debug}
        logman = LoggerManager(__name__, prefix=self.prefix, **kwargs)
        self.logger = logman.get_logger()

    def read_012(self) -> None:
        """Read 012-encoded comma-delimited file.

        This method reads a 012-encoded comma-delimited file and stores the data in the GenotypeEncoder object.

        Raises:
            ValueError: Sequences differ in length.
        """
        self.logger.info(f"Reading 012-encoded file: {self.filename}...")

        self._check_filetype("encoded")
        snp_data, num_snps = [], []

        with open(self.filename, "r") as fin:
            num_inds = 0
            for line in fin:
                line = line.strip()
                if not line:
                    continue
                cols = line.split(",")
                inds = cols[0]
                snps = cols[1:]
                num_snps.append(len(snps))
                num_inds += 1
                snp_data.append(snps)
                self.samples.append(inds)

        if len(list(set(num_snps))) > 1:
            msg = "All sequences must be the same length. At least one sequence differs in length from the others"
            self.logger.error(msg)
            raise ValueError(msg)

        miss_vals = ["NA", np.nan, pd.NA, "N", ".", "-", "?"]
        df = pd.DataFrame(snp_data)
        df = df.replace(miss_vals, ["-9"] * len(miss_vals)).astype(int)

        # Decodes 012 and converts to self.snp_data
        self.genotypes_012 = df

        self.ref = None
        self.alt = None

        self.logger.info("012 file successfully loaded!")
        self.logger.info(
            f"\nFound:\n"
            f"\t{self.num_inds} individuals...\n"
            f"\t{self.num_snps} SNPs...\n"
        )

    def convert_012(self, snps: List[List[str]]) -> List[List[int]]:
        """Encode IUPAC nucleotides as 0 (reference), 1 (heterozygous), and 2 (alternate) alleles.

        This method encodes IUPAC nucleotides as 0 (reference), 1 (heterozygous), and 2 (alternate) alleles.

        Args:
            snps (List[List[str]]): 2D list of genotypes of shape (n_samples, n_sites).

        Returns:
            List[List[int]]: Encoded 012 genotypes.

        Warning:
            Monomorphic sites are detected and encoded as 0 (reference).

            Non-biallelic sites are detected and forced to be bi-allelic.

            Sites with all missing data are detected and excluded from the alignment.
        """
        snps_012 = []
        new_snps = []
        monomorphic_sites = []
        non_biallelic_sites = []
        all_missing = []

        for i in range(0, len(snps)):
            new_snps.append([])

        for j in range(0, len(snps[0])):
            loc = []
            for i in range(0, len(snps)):
                loc.append(snps[i][j].upper())

            if all(x == "N" for x in loc):
                all_missing.append(j)
                continue
            num_alleles = sequence_tools.count_alleles(loc)
            if num_alleles != 2:
                # If monomorphic
                if num_alleles < 2:
                    monomorphic_sites.append(j)
                    try:
                        ref = list(
                            map(
                                sequence_tools.get_major_allele,
                                loc,
                                [False for x in loc],
                            )
                        )
                        ref = str(ref[0])
                    except IndexError:
                        ref = list(
                            map(
                                sequence_tools.get_major_allele,
                                loc,
                                [False for x in loc],
                            )
                        )
                    alt = None

                    for i in range(0, len(snps)):
                        if loc[i] in ["-", "-9", "N"]:
                            new_snps[i].append(-9)

                        elif loc[i] == ref:
                            new_snps[i].append(0)

                        else:
                            new_snps[i].append(1)

                # If >2 alleles
                elif num_alleles > 2:
                    non_biallelic_sites.append(j)
                    all_alleles = sequence_tools.get_major_allele(loc)
                    all_alleles = [str(x[0]) for x in all_alleles]
                    ref = all_alleles.pop(0)
                    alt = all_alleles.pop(0)
                    others = all_alleles

                    for i in range(0, len(snps)):
                        if loc[i] in ["-", "-9", "N"]:
                            new_snps[i].append(-9)

                        elif loc[i] == ref:
                            new_snps[i].append(0)

                        elif loc[i] == alt:
                            new_snps[i].append(2)

                        # Force biallelic
                        elif loc[i] in others:
                            new_snps[i].append(2)

                        else:
                            new_snps[i].append(1)

            else:
                ref, alt = sequence_tools.get_major_allele(loc)
                ref = str(ref)
                alt = str(alt)

                for i in range(0, len(snps)):
                    if loc[i] in ["-", "-9", "N"]:
                        new_snps[i].append(-9)

                    elif loc[i] == ref:
                        new_snps[i].append(0)

                    elif loc[i] == alt:
                        new_snps[i].append(2)

                    else:
                        new_snps[i].append(1)

        outdir = Path(f"{self.prefix}_output", "gtdata", "logs")
        outdir.mkdir(exist_ok=True, parents=True)
        if monomorphic_sites:
            # TODO: Check here if column is all missing.
            # TODO: What to do in this case? Error out?
            outfile = outdir / "monomorphic_sites.txt"
            with open(outfile, "w") as fout:
                mono_sites = [str(x) for x in monomorphic_sites]
                fout.write(",".join(mono_sites))

            self.logger.warning(
                f"Monomorphic sites detected. You can check the locus indices in the following log file: {outfile}"
            )

        if non_biallelic_sites:
            outfile = outdir / "non_biallelic_sites.txt"
            with open(outfile, "w") as fout:
                nba = [str(x) for x in non_biallelic_sites]
                fout.write(",".join(nba))

            self.logger.warning(
                f"SNP column indices listed in the log file {outfile} had >2 alleles and was forced to be bi-allelic. If that is not desired, please fix or remove the column and re-run."
            )

        if all_missing:
            outfile = outdir / "all_missing_sites.txt"
            with open(outfile, "w") as fout:
                ",".join([str(x) for x in all_missing])

            self.logger.warning(
                f"SNP column indices found in the log file {outfile} had all missing data and were excluded from the alignment."
            )

        snps_012 = [s for s in new_snps]
        return snps_012

    def convert_onehot(
        self,
        snp_data: Union[np.ndarray, List[List[int]]],
        encodings_dict: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        """Convert input data to one-hot encoded format.

        This method converts input data to one-hot encoded format.

        Args:
            snp_data (Union[np.ndarray, List[List[int]]]): Input 012-encoded data of shape (n_samples, n_SNPs).

            encodings_dict (Optional[Dict[str, int]]): Encodings to convert structure to phylip format. Defaults to None.

        Returns:
            np.ndarray: One-hot encoded data.

        Note:
            If the data file type is "phylip" and `encodings_dict` is not provided, default encodings for nucleotides are used.

            If the data file type is "structure1row" or "structure2row" and `encodings_dict` is not provided, default encodings for alleles are used.

            Otherwise, if `encodings_dict` is provided, it will be used for conversion.

        Warning:
            If the data file type is "structure1row" or "structure2row" and `encodings_dict` is not provided, default encodings for alleles are used.

            If the data file type is "phylip" and `encodings_dict` is not provided, default encodings for nucleotides are used.

            If the data file type is "structure" and `encodings_dict` is not provided, default encodings for alleles are used.
        """

        if encodings_dict is None:
            onehot_dict = misc.get_onehot_dict()
        else:
            if isinstance(snp_data, np.ndarray):
                snp_data = snp_data.tolist()
            onehot_dict = encodings_dict
        onehot_outer_list = list()

        n_rows = len(self.samples) if encodings_dict is None else len(snp_data)

        for i in range(n_rows):
            onehot_list = list()
            for j in range(len(snp_data[0])):
                onehot_list.append(onehot_dict[snp_data[i][j]])
            onehot_outer_list.append(onehot_list)

        return np.array(onehot_outer_list)

    def inverse_onehot(
        self,
        onehot_data: Union[np.ndarray, List[List[float]]],
        encodings_dict: Optional[Dict[str, List[float]]] = None,
    ) -> np.ndarray:
        """Convert one-hot encoded data back to original format.

        Args:
            onehot_data (Union[np.ndarray, List[List[float]]]): Input one-hot encoded data of shape (n_samples, n_SNPs).

            encodings_dict (Optional[Dict[str, List[float]]]): Encodings to convert from one-hot encoding to original format. Defaults to None.

        Returns:
            np.ndarray: Original format data.

        Note:
            If the data file type is "phylip" or "vcf" and `encodings_dict` is not provided, default encodings based on IUPAC codes are used.

            If the data file type is "structure" and `encodings_dict` is not provided, default encodings for alleles are used.

            Otherwise, if `encodings_dict` is provided, it will be used for conversion.

            If the input data is a numpy array, it will be converted to a list of lists before decoding.
        """

        onehot_dict = (
            misc.get_onehot_dict() if encodings_dict is None else encodings_dict
        )

        # Create inverse dictionary (from list to key)
        inverse_onehot_dict = {tuple(v): k for k, v in onehot_dict.items()}

        if isinstance(onehot_data, np.ndarray):
            onehot_data = onehot_data.tolist()

        decoded_outer_list = []

        for i in range(len(onehot_data)):
            decoded_list = []
            for j in range(len(onehot_data[0])):
                # Look up original key using one-hot encoded list
                decoded_list.append(inverse_onehot_dict[tuple(onehot_data[i][j])])
            decoded_outer_list.append(decoded_list)

        return np.array(decoded_outer_list)

    def convert_int_iupac(
        self,
        snp_data: Union[np.ndarray, List[List[int]]],
        encodings_dict: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        """Convert input data to integer-encoded format (0-9) based on IUPAC codes.

        This method converts input data to integer-encoded format (0-9) based on IUPAC codes. The integer encoding is as follows: A=0, T=1, G=2, C=3, W=4, R=5, M=6, K=7, Y=8, S=9, N=-9.

        Args:
            snp_data (numpy.ndarray of shape (n_samples, n_SNPs) or List[List[int]]): Input 012-encoded data.

            encodings_dict (Dict[str, int] or None): Encodings to convert structure to phylip format.

        Returns:
            numpy.ndarray: Integer-encoded data.

        Note:
            If the data file type is "phylip" or "vcf" and ``encodings_dict`` is not provided, default encodings based on IUPAC codes are used.

            If the data file type is "structure" and ``encodings_dict`` is not provided, default encodings for alleles are used.

            Otherwise, if ``encodings_dict`` is provided, it will be used for conversion.
        """

        if encodings_dict is None:
            int_iupac_dict = misc.get_int_iupac_dict()
        else:
            if isinstance(snp_data, np.ndarray):
                snp_data = snp_data.tolist()

            int_iupac_dict = encodings_dict

        outer_list = list()

        n_rows = len(self.samples) if encodings_dict is None else len(snp_data)

        for i in range(n_rows):
            int_iupac = list()
            for j in range(len(snp_data[0])):
                int_iupac.append(int_iupac_dict[snp_data[i][j]])
            outer_list.append(int_iupac)

        return np.array(outer_list)

    def inverse_int_iupac(
        self,
        int_encoded_data: Union[np.ndarray, List[List[int]]],
        encodings_dict: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        """Convert integer-encoded data back to original format.

        This method converts integer-encoded data back to the original format based on IUPAC codes. The integer encoding is as follows: A=0, T=1, G=2, C=3, W=4, R=5, M=6, K=7, Y=8, S=9, N=-9.

        Args:
            int_encoded_data (numpy.ndarray of shape (n_samples, n_SNPs) or List[List[int]]): Input integer-encoded data.

            encodings_dict (Dict[str, int] or None): Encodings to convert from integer encoding to original format.

        Returns:
            numpy.ndarray: Original format data.

        Note:
            If the data file type is "phylip" or "vcf" and `encodings_dict` is not provided, default encodings based on IUPAC codes are used.

            If the data file type is "structure" and `encodings_dict` is not provided, default encodings for alleles are used.

            Otherwise, if `encodings_dict` is provided, it will be used for conversion
        """

        int_encodings_dict = (
            misc.get_int_iupac_dict() if encodings_dict is None else encodings_dict
        )

        # Create inverse dictionary (from integer to key)
        inverse_int_encodings_dict = {v: k for k, v in int_encodings_dict.items()}

        if isinstance(int_encoded_data, np.ndarray):
            int_encoded_data = int_encoded_data.tolist()

        decoded_outer_list = []

        for i in range(len(int_encoded_data)):
            decoded_list = []
            for j in range(len(int_encoded_data[0])):
                # Look up original key using integer encoding
                decoded_list.append(inverse_int_encodings_dict[int_encoded_data[i][j]])
            decoded_outer_list.append(decoded_list)

        return np.array(decoded_outer_list)

    def decode_012(
        self,
        X: Union[np.ndarray, pd.DataFrame, List[List[int]]],
        write_output: bool = True,
        is_nuc: bool = False,
    ):
        """Decode 012-encoded or 0-9 integer-encoded imputed data to STRUCTURE or PHYLIP format.

        This method decodes 012-encoded or 0-9 integer-encoded imputed data to IUPAC format. The decoded data can be saved to a file or returned as a DataFrame.

        Args:
            X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Imputed data to decode, encoded as 012 or 0-9 integers.

            write_output (bool, optional): If True, save the decoded output to a file. If False, return the decoded data as a DataFrame. Defaults to True.

            is_nuc (bool, optional): Whether the encoding is based on nucleotides instead of 012. Defaults to False.

        Returns:
            str or pandas.DataFrame: If write_output is True, returns the filename where the imputed data was written. If write_output is False, returns the decoded data as a DataFrame.

        TODO:
            - Check if VAE still uses IUPAC encodings.
        """
        if isinstance(X, pd.DataFrame):
            df = X.copy()
        elif isinstance(X, (np.ndarray, list)):
            df = pd.DataFrame(X)

        nuc = {
            "A/A": "A",
            "T/T": "T",
            "G/G": "G",
            "C/C": "C",
            "A/G": "R",
            "G/A": "R",
            "C/T": "Y",
            "T/C": "Y",
            "G/C": "S",
            "C/G": "S",
            "A/T": "W",
            "T/A": "W",
            "G/T": "K",
            "T/G": "K",
            "A/C": "M",
            "C/A": "M",
            "N/N": "N",
        }

        ft = self.filetype.lower()

        is_phylip = False
        if ft == "phylip" or ft == "vcf":
            is_phylip = True

        df_decoded = df.copy()
        df_decoded = df.copy().astype(object)

        # TODO: Check if VAE still uses IUPAC encodings.
        # VAE uses [A,T,G,C] encodings. The other NN methods use
        # [0,1,2] encodings.
        if is_nuc:
            classes_int = range(10)
            classes_string = [str(x) for x in classes_int]
            if is_phylip:
                gt = ["A", "T", "G", "C", "W", "R", "M", "K", "Y", "S", "N"]
            else:
                gt = [
                    "1/1",
                    "2/2",
                    "3/3",
                    "4/4",
                    "1/2",
                    "1/3",
                    "1/4",
                    "2/3",
                    "2/4",
                    "3/4",
                    "-9/-9",
                ]
            d = dict(zip(classes_int, gt))
            dstr = dict(zip(classes_string, gt))
            d.update(dstr)
            dreplace = {col: d for col in list(df.columns)}

        else:
            dreplace = dict()
            for col, ref, alt in zip(df.columns, self._ref, self._alt):
                # if site is monomorphic, set alt and ref state the
                # same
                if alt is None:
                    alt = ref
                ref2 = f"{ref}/{ref}"
                alt2 = f"{alt}/{alt}"
                het2 = f"{ref}/{alt}"

                if is_phylip:
                    ref2 = nuc[ref2]
                    alt2 = nuc[alt2]
                    het2 = nuc[het2]

                d = {
                    "0": ref2,
                    0: ref2,
                    "1": het2,
                    1: het2,
                    "2": alt2,
                    2: alt2,
                    "-9": "N",
                    -9: "N",
                }
                dreplace[col] = d

        df_decoded = df_decoded.replace(dreplace)

        if write_output:
            outfile = Path(f"{self.prefix}_output", "gtdata", "alignments", "012")

        if ft.startswith("structure"):
            if ft.startswith("structure2row"):
                for col in df_decoded.columns:
                    df_decoded[col] = (
                        df_decoded[col]
                        .str.split("/")
                        .apply(lambda x: list(map(int, x)))
                    )

                df_decoded.insert(0, "sampleID", self._samples)
                df_decoded.insert(1, "popID", self._populations)

                # Transform each element to a separate row.
                df_decoded = (
                    df_decoded.set_index(["sampleID", "popID"])
                    .apply(pd.Series.explode)
                    .reset_index()
                )

            elif ft.startswith("structure1row"):
                df_decoded = pd.concat(
                    [
                        df_decoded[c]
                        .astype(str)
                        .str.split("/", expand=True)
                        .add_prefix(f"{c}_")
                        for c in df_decoded.columns
                    ],
                    axis=1,
                )

            elif ft == "structure":
                for col in df_decoded.columns:
                    df_decoded[col] = (
                        df_decoded[col]
                        .str.split("/")
                        .apply(lambda x: list(map(int, x)))
                    )

                df_decoded.insert(0, "sampleID", self._samples)
                df_decoded.insert(1, "popID", self._populations)

                # Transform each element to a separate row.
                df_decoded = (
                    df_decoded.set_index(["sampleID", "popID"])
                    .apply(pd.Series.explode)
                    .reset_index()
                )

            if write_output:
                of = outfile.with_suffix(".str")
                df_decoded.insert(0, "sampleID", self._samples)
                df_decoded.insert(1, "popID", self._populations)

                df_decoded.to_csv(of, sep="\t", header=False, index=False)

        elif ft.startswith("phylip"):
            if write_output:
                of = outfile.with_suffix(".phy")
                header = f"{self.num_inds} {self.num_snps}\n"
                with open(of, "w") as fout:
                    fout.write(header)

                lst_decoded = df_decoded.values.tolist()

                with open(of, "a") as fout:
                    for sample, row in zip(self._samples, lst_decoded):
                        seqs = "".join([str(x) for x in row])
                        fout.write(f"{sample}\t{seqs}\n")

        if write_output:
            return of
        return df_decoded.values.tolist()

    @property
    def genotypes_012(
        self,
    ) -> Union[List[List[int]], np.ndarray, pd.DataFrame]:
        """Encoded 012 genotypes as a 2D list, numpy array, or pandas DataFrame.

        This method encodes genotypes as 0 (reference), 1 (heterozygous), and 2 (alternate) alleles. The encoded genotypes are returned as a 2D list, numpy array, or pandas DataFrame.

        Returns:
            List[List[int]], np.ndarray, or pd.DataFrame: encoded 012 genotypes.

        Example:
            >>> gd = VCFReader(filename="example_data/vcf_files/phylogen_subset14K_sorted.vcf.gz", popmapfile="example_data/popmaps/phylogen_nomx.popmap", force_popmap=True, chunk_size=5000, verbose=False)
            >>> ge = GenotypeEncoder(gd)
            >>> gt012 = ge.genotypes_012
            >>> print(gt012)
            [["0", "1", "2"], ["0", "1", "2"], ["0", "1", "2"]]
        """
        g012 = self.convert_012(self.snp_data)
        self.logger.debug(f"Genotypes 012: {g012}")
        return g012

    @genotypes_012.setter
    def genotypes_012(self, value: np.ndarray) -> None:
        """Set the 012 genotypes. They will be decoded back to a 2D list of genotypes as ``snp_data`` object.

        012-encoded genotypes are returned as a 2D numpy array of shape (n_samples, n_sites). The encoding is as follows: 0=reference, 1=heterozygous, 2=alternate allele.

        Args:
            value (np.ndarray): 2D numpy array with 012-encoded genotypes.
        """
        self.snp_data = self.decode_012(value, write_output=False)
        self.logger.debug(f"Decoded 012 genotypes: {self.snp_data}")

    @property
    def genotypes_onehot(self) -> np.ndarray:
        """One-hot encoded snps format of shape (n_samples, n_loci, 4).

        One-hot encoded genotypes are returned as a 3D numpy array of shape (n_samples, n_loci, 4).  The one-hot encoding is as follows: A=[1, 0, 0, 0], T=[0, 1, 0, 0], G=[0, 0, 1, 0], C=[0, 0, 0, 1]. Missing values are encoded as [0, 0, 0, 0]. The one-hot encoding is based on the IUPAC ambiguity codes. Heterozygous sites are encoded as 0.5 for each allele.

        Returns:
            numpy.ndarray: One-hot encoded numpy array of shape (n_samples, n_loci, 4).
        """
        gohe = self.convert_onehot(self.snp_data)
        self.logger.debug(f"Genotypes one-hot encoded: {gohe}")
        return gohe

    @genotypes_onehot.setter
    def genotypes_onehot(
        self, value: Union[np.ndarray, List[List[List[int]]], pd.DataFrame]
    ) -> None:
        """Set the onehot-encoded genotypes. They will be decoded back to a 2D list of IUPAC genotypes as ``snp_data``.

        One-hot encoded genotypes are returned as a 3D numpy array of shape (n_samples, n_loci, 4).  The one-hot encoding is as follows: A=[1, 0, 0, 0], T=[0, 1, 0, 0], G=[0, 0, 1, 0], C=[0, 0, 0, 1]. Missing values are encoded as [0, 0, 0, 0]. The one-hot encoding is based on the IUPAC ambiguity codes. Heterozygous sites are encoded as 0.5 for each allele.

        Args:
            value (Union[np.ndarray, List[List[List[int]]], pd.DataFrame]): 3D numpy array with one-hot encoded genotypes.

        Raises:
            TypeError: If `value` is not of type pd.DataFrame, np.ndarray, or list.
        """
        if isinstance(value, pd.DataFrame):
            X = value.to_numpy()
        elif isinstance(value, list):
            X = np.array(value)
        elif isinstance(value, np.ndarray):
            X = value
        else:
            msg = f"genotypes_onehot must be of type pd.DataFrame, np.ndarray, or list, but got {type(value)}"
            self.logger.error(msg)
            raise TypeError(msg)

        Xt = self.inverse_onehot(X)
        self.snp_data = Xt
        self.logger.debug(f"Decoded one-hot genotypes: {Xt}")

    @property
    def genotypes_int(self) -> np.ndarray:
        """Integer-encoded (0-9 including IUPAC characters) snps format.

        Integer-encoded genotypes are returned as a 2D numpy array of shape (n_samples, n_sites). The integer encoding is as follows: A=0, T=1, G=2, C=3, W=4, R=5, M=6, K=7, Y=8, S=9, N=-9. Missing values are encoded as -9.

        Returns:
            numpy.ndarray: 2D array of shape (n_samples, n_sites), integer-encoded from 0-9 with IUPAC characters.
        """
        gint = self.convert_int_iupac(self.snp_data)
        self.logger.debug(f"Genotypes integer-encoded: {gint}")
        return gint

    @genotypes_int.setter
    def genotypes_int(
        self, value: Union[pd.DataFrame, np.ndarray, List[List[int]], Any]
    ) -> None:
        """Set the integer-encoded (0-9) genotypes. They will be decoded back to a 2D list of IUPAC genotypes as a ``snp_data`` object.

        Integer-encoded genotypes are returned as a 2D numpy array of shape (n_samples, n_sites). The integer encoding is as follows: A=0, T=1, G=2, C=3, W=4, R=5, M=6, K=7, Y=8, S=9, N=-9. Missing values are encoded as -9.
        """
        if isinstance(value, pd.DataFrame):
            X = value.to_numpy()
        elif isinstance(value, list):
            X = np.array(value)
        elif isinstance(value, np.ndarray):
            X = value
        else:
            msg = f"genotypes_int must be of type pd.DataFrame, np.ndarray, or list, but got {type(value)}"
            self.logger.error(msg)
            raise TypeError(msg)

        Xt = self.inverse_int_iupac(X)
        self.snp_data = Xt
        self.logger.debug(f"Decoded integer-encoded genotypes: {Xt}")



import csv
import unittest
from pathlib import Path

from snpio import NRemover2, VCFReader


class TestNRemover2(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Create a small test VCF file with controlled data for all filter tests
        cls.test_vcf_file = "test_data.vcf"
        if not Path(cls.test_vcf_file).exists():
            with open(cls.test_vcf_file, "w", newline="") as vcf:
                writer = csv.writer(vcf, delimiter="\t", quoting=csv.QUOTE_MINIMAL)

                # Write the VCF header
                writer.writerow(["##fileformat=VCFv4.2"])
                writer.writerow(
                    ["##FORMAT=<ID=GT,Number=1,Type=String,Description='Genotype'>"]
                )
                writer.writerow(
                    [
                        "#CHROM",
                        "POS",
                        "ID",
                        "REF",
                        "ALT",
                        "QUAL",
                        "FILTER",
                        "INFO",
                        "FORMAT",
                        "Sample1",
                        "Sample2",
                        "Sample3",
                    ]
                )

                # Write the VCF data rows (each field strictly tab-delimited)
                writer.writerow(
                    [
                        "NW123.1",
                        100,
                        ".",
                        "A",
                        "T",
                        ".",
                        "PASS",
                        ".",
                        "GT",
                        "0/0",
                        "0/1",
                        "1/1",
                    ]
                )
                writer.writerow(
                    [
                        "NW123.1",
                        200,
                        ".",
                        "G",
                        "A",
                        ".",
                        "PASS",
                        ".",
                        "GT",
                        "0/1",
                        "./.",
                        "0/1",
                    ]
                )
                writer.writerow(
                    [
                        "XM123.1",
                        100,
                        ".",
                        "C",
                        ".",
                        ".",
                        "PASS",
                        ".",
                        "GT",
                        "0/0",
                        "0/0",
                        "0/0",
                    ]
                )
                writer.writerow(
                    [
                        "XM123.1",
                        201,
                        ".",
                        "T",
                        "G",
                        ".",
                        "PASS",
                        ".",
                        "GT",
                        "1/1",
                        "1/1",
                        "0/1",
                    ]
                )
                writer.writerow(
                    [
                        "XM123.1",
                        305,
                        ".",
                        "C",
                        "G,A",
                        ".",
                        "PASS",
                        ".",
                        "GT",
                        "0/2",
                        "0/1",
                        "./.",
                    ]
                )

    def setUp(self):
        # Initialize the VCFReader with the test VCF file
        self.vcf_reader = VCFReader(
            filename=self.test_vcf_file,
            popmapfile=None,
            force_popmap=False,
            chunk_size=100,
        )

        # Initialize NRemover2 with the VCFReader instance
        self.n_remover = NRemover2(self.vcf_reader)

    def test_filter_missing_sample(self):
        # Test filter_missing_sample with a threshold of 0.19 (19% missing allowed)
        filtered_data = self.n_remover.filter_missing_sample(0.19).resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.sample_indices) if keep
        ]
        expected_indices = [0]  # Indices of retained samples
        self.assertEqual(retained_indices, expected_indices)

    def test_filter_missing(self):
        # Test filter_missing loci with a threshold of 0.3 (30% missing allowed)
        filtered_data = self.n_remover.filter_missing(0.3).resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.loci_indices) if keep
        ]
        expected_indices = [0, 2, 3]  # Indices of retained loci
        self.assertEqual(retained_indices, expected_indices)

    def test_filter_mac(self):
        # Test filter_mac to keep loci with MAC >= 2
        filtered_data = self.n_remover.filter_mac(2).resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.loci_indices) if keep
        ]
        expected_indices = [0]  # Indices of retained loci
        self.assertEqual(retained_indices, expected_indices)

    def test_filter_monomorphic(self):
        # Test filter_monomorphic to remove monomorphic loci
        filtered_data = self.n_remover.filter_monomorphic().resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.loci_indices) if keep
        ]
        expected_indices = [0, 3, 4]  # Locus 2 filtered out as monomorphic
        self.assertEqual(retained_indices, expected_indices)

    def test_filter_singletons(self):
        # Test filter_singletons to remove loci that only have one
        # non-reference allele
        filtered_data = self.n_remover.filter_singletons().resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.loci_indices) if keep
        ]
        expected_indices = [1, 2]  # Indices of retained loci
        self.assertEqual(retained_indices, expected_indices)

    def test_filter_biallelic(self):
        # Test filter_biallelic to keep only loci with two alleles
        filtered_data = self.n_remover.filter_biallelic().resolve()
        retained_indices = [
            i for i, keep in enumerate(filtered_data.loci_indices) if keep
        ]
        expected_indices = [0, 1, 3]  # Locus 2 monomorphic; locus 4 multiallelic
        self.assertEqual(retained_indices, expected_indices)

    @classmethod
    def tearDownClass(cls):
        # Clean up by deleting the test VCF file
        if Path(cls.test_vcf_file).exists():
            Path(cls.test_vcf_file).unlink()

        if Path(cls.test_vcf_file + ".tbi").exists():
            Path(cls.test_vcf_file + ".tbi").unlink()


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional


class UnsupportedFileTypeError(Exception):
    """Custom exception to check if the given file type is supported.

    Attributes:
        filetype (str): The type of the file being checked.
        message (str): Explanation of the error.
        supported_types (list): List of supported file types.
    """

    def __init__(
        self, filetype: str, supported_types: Optional[List[str]] = None
    ) -> None:
        """Initialize UnsupportedFileTypeException.

        This exception is raised when the given file type is not supported.

        Args:
            filetype (str): The type of the file being checked.
            supported_types (list, optional): List of supported file types.
                Defaults to ["vcf", "phylip", "structure"].
        """
        if supported_types is None:
            supported_types = ["vcf", "phylip", "structure"]

        self.filetype = filetype
        self.supported_types = supported_types
        self.message = f"Alignment file type '{self.filetype}' is not supported. Supported types are {', '.join(self.supported_types)}."
        super().__init__(self.message)

    def __str__(self) -> str:
        """String representation of the exception.

        Returns:
            str: The message associated with the exception.
        """
        return self.message


class AlignmentError(Exception):
    """Base class for exceptions related to alignment file handling."""

    pass


class AlignmentFormatError(AlignmentError):
    """Raised when an alignment file has an invalid format."""

    def __init__(self, message: str) -> None:
        """Initialize AlignmentFormatError.

        Args:
            message (str): The error message.
        """
        super().__init__(message)


class SequenceLengthError(AlignmentFormatError):
    """Raised when sequences in the alignment file have unequal lengths."""

    def __init__(self, sample_name: str) -> None:
        """Initialize SequenceLengthError.

        Args:
            sample_name (str): The name of the sample where the error occurred.
        """
        msg = f"Sequences have unequal lengths. Error at sample: {sample_name}"
        super().__init__(msg)


class PhylipAlignmentSampleMismatch(AlignmentFormatError):
    """Raised when samples in the alignment file do not match the expected number of samples."""

    def __init__(
        self, header_samples: int, sample_list_len: int, snp_data_len: int
    ) -> None:
        """Initialize PhylipAlignmentSampleMismatch.

        Args:
            header_samples (int): The number of samples in the header.
            sample_list_len (int): The number of samples in the sample list.
            snp_data_len (int): The number of genotypes in the SNP data.
        """
        super().__init__(
            f"Unexpected number of samples encountered. Expected {header_samples}, but got: {sample_list_len} samples across {snp_data_len} genotypes."
        )


class StructureAlignmentSampleMismatch(AlignmentFormatError):
    """Raised when samples in the alignment file do not match the expected number of samples."""

    def __init__(self, sample_list_len: int, snp_data_len: int) -> None:
        """Initialize StructureAlignmentSampleMismatch.

        Args:
            sample_list_len (int): The number of samples in the sample list.
            snp_data_len (int): The number of genotypes in the SNP data.
        """
        super().__init__(
            f"Unexpected number of samples encountered. Expected {sample_list_len}, but got: {snp_data_len} genotype rows."
        )


class AlignmentFileNotFoundError(AlignmentError):
    """Raised when an alignment file is not found."""

    def __init__(self, filename: str) -> None:
        """Initialize AlignmentFileNotFoundError.

        Args:
            filename (str): The name of the alignment file that was not found.
        """
        super().__init__(f"Alignment file '{filename}' not found.")


class NoValidAllelesError(Exception):
    """Custom exception raised when no valid alleles are found in a SNP column."""

    def __init__(self, column_index: int) -> None:
        """Initialize NoValidAllelesError.

        Args:
            column_index (int): The index of the column where no valid alleles were found.
        """
        super().__init__(f"No valid alleles found in column {column_index}")
        self.column_index = column_index



import unittest
from unittest.mock import MagicMock, patch, call

from run_snpio import main
from snpio import NRemover2, Plotting, VCFReader


class TestRunSnpio(unittest.TestCase):

    @patch("run_snpio.VCFReader")
    @patch("run_snpio.Plotting")
    @patch("run_snpio.NRemover2")
    def test_main(self, MockNRemover2, MockPlotting, MockVCFReader):
        # Mock the VCFReader instance
        mock_vcf_reader = MockVCFReader.return_value
        mock_vcf_reader.missingness_reports = MagicMock()

        # Mock the Plotting instance
        mock_plotting = MockPlotting.return_value
        mock_plotting.run_pca = MagicMock(return_value=(MagicMock(), MagicMock()))

        # Mock the NRemover2 instance
        mock_nremover2 = MockNRemover2.return_value
        mock_nremover2.filter_missing_sample.return_value = mock_nremover2
        mock_nremover2.filter_missing.return_value = mock_nremover2
        mock_nremover2.filter_missing_pop.return_value = mock_nremover2
        mock_nremover2.filter_mac.return_value = mock_nremover2
        mock_nremover2.filter_monomorphic.return_value = mock_nremover2
        mock_nremover2.filter_singletons.return_value = mock_nremover2
        mock_nremover2.filter_biallelic.return_value = mock_nremover2
        mock_nremover2.resolve.return_value = mock_nremover2
        mock_nremover2.plot_sankey_filtering_report = MagicMock()

        # Mock the filtered genotype data
        mock_gd_filt = mock_nremover2
        mock_gd_filt.missingness_reports = MagicMock()
        mock_gd_filt.write_vcf = MagicMock()

        # Run the main function
        main()

        # Assertions to ensure the methods were called
        MockVCFReader.assert_called_once_with(
            filename="example_data/vcf_files/phylogen_subset14K_sorted.vcf.gz",
            popmapfile="example_data/popmaps/phylogen_nomx.popmap",
            force_popmap=True,
            chunk_size=5000,
        )
        mock_vcf_reader.missingness_reports.assert_called_once()

        # Instead of assert_has_calls, use assert_any_call for individual calls
        MockPlotting.assert_any_call(genotype_data=mock_vcf_reader)
        MockPlotting.assert_any_call(genotype_data=mock_gd_filt)

        mock_plotting.run_pca.assert_called()  # Check if run_pca was called at least once
        MockNRemover2.assert_called_with(mock_vcf_reader)
        mock_nremover2.filter_missing_sample.assert_called_once_with(0.75)
        mock_nremover2.filter_missing.assert_called_once_with(0.75)
        mock_nremover2.filter_missing_pop.assert_called_once_with(0.75)
        mock_nremover2.filter_mac.assert_called_once_with(2)
        mock_nremover2.filter_monomorphic.assert_called_once_with(
            exclude_heterozygous=False
        )
        mock_nremover2.filter_singletons.assert_called_once_with(
            exclude_heterozygous=False
        )
        mock_nremover2.filter_biallelic.assert_called_once_with(
            exclude_heterozygous=False
        )
        mock_nremover2.resolve.assert_called_once()
        mock_nremover2.plot_sankey_filtering_report.assert_called_once()

        mock_gd_filt.missingness_reports.assert_called_once_with(prefix="filtered")
        mock_gd_filt.write_vcf.assert_called_once_with(
            "example_data/vcf_files/nremover_test.vcf"
        )


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List, Union

import numpy as np
import pandas as pd

from snpio.utils.logging import LoggerManager

logman = LoggerManager(__name__, verbose=False, debug=False)
logger = logman.get_logger()


def get_gt2iupac() -> Dict[str, str]:
    """Get a dictionary of genotype to IUPAC ambiguity codes."""
    return {
        "1/1": "A",
        "2/2": "C",
        "3/3": "G",
        "4/4": "T",
        "1/2": "M",  # A/C
        "1/3": "R",  # A/G
        "1/4": "W",  # A/T
        "2/3": "S",  # C/G
        "2/4": "Y",  # C/T
        "3/4": "K",  # G/T
        "-9/-9": "N",  # Missing data
        "-1/-1": "N",  # Missing data
    }


def get_iupac2gt() -> Dict[str, str]:
    """Get a dictionary of IUPAC ambiguity codes to genotype."""
    return {
        "A": "1/1",
        "C": "2/2",
        "G": "3/3",
        "T": "4/4",
        "M": "1/2",  # A/C
        "R": "1/3",  # A/G
        "W": "1/4",  # A/T
        "S": "2/3",  # C/G
        "Y": "2/4",  # C/T
        "K": "3/4",  # G/T
        "N": "-9/-9",  # Missing data
    }


def get_int_iupac_dict() -> Dict[str, int]:
    """Get a dictionary of IUPAC ambiguity codes to integers."""
    int_iupac_dict = {
        "A": 0,
        "T": 1,
        "G": 2,
        "C": 3,
        "W": 4,
        "R": 5,
        "M": 6,
        "K": 7,
        "Y": 8,
        "S": 9,
        "N": -9,
    }

    return int_iupac_dict


def get_onehot_dict() -> Dict[str, List[float]]:
    """Get a dictionary of IUPAC ambiguity codes to one-hot encoded vectors."""
    onehot_dict = {
        "A": [1.0, 0.0, 0.0, 0.0],
        "T": [0.0, 1.0, 0.0, 0.0],
        "G": [0.0, 0.0, 1.0, 0.0],
        "C": [0.0, 0.0, 0.0, 1.0],
        "N": [0.0, 0.0, 0.0, 0.0],
        "W": [0.5, 0.5, 0.0, 0.0],
        "R": [0.5, 0.0, 0.5, 0.0],
        "M": [0.5, 0.0, 0.0, 0.5],
        "K": [0.0, 0.5, 0.5, 0.0],
        "Y": [0.0, 0.5, 0.0, 0.5],
        "S": [0.0, 0.0, 0.5, 0.5],
        "N": [0.0, 0.0, 0.0, 0.0],
    }

    return onehot_dict


def validate_input_type(
    X: Union[np.ndarray, pd.DataFrame, List[List[int]]], return_type: str = "array"
) -> Union[np.ndarray, pd.DataFrame, List[List[int]]]:
    """Validates the input type and returns it as a specified type.

    This function checks if the input `X` is a pandas DataFrame, numpy array, or a list of lists. It then converts `X` to the specified `return_type` and returns it.

    Args:
        X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): The input data to validate and convert.

        return_type (str, optional): The type of the returned object. Supported options include: "df" (DataFrame), "array" (numpy array), and "list". Defaults to "array".

    Returns:
        pandas.DataFrame, numpy.ndarray, or List[List[int]]: The input data converted to the desired return type.

    Raises:
        TypeError: If `X` is not of type pandas.DataFrame, numpy.ndarray, or List[List[int]].

        ValueError: If an unsupported `return_type` is provided. Supported types are "df", "array", and "list".

    Example:
        >>> X = [[1, 2, 3], [4, 5, 6]]
        >>> print(validate_input_type(X, "df"))  4
        >>> # Outputs: a DataFrame with the data from `X`.
    """
    if not isinstance(X, (pd.DataFrame, np.ndarray, list)):
        raise TypeError(
            f"X must be of type pandas.DataFrame, numpy.ndarray, "
            f"or List[List[int]], but got {type(X)}"
        )

    if return_type == "array":
        if isinstance(X, pd.DataFrame):
            return X.to_numpy()
        elif isinstance(X, list):
            return np.array(X)
        elif isinstance(X, np.ndarray):
            return X.copy()

    elif return_type == "df":
        if isinstance(X, pd.DataFrame):
            return X.copy()
        elif isinstance(X, (np.ndarray, list)):
            return pd.DataFrame(X)

    elif return_type == "list":
        if isinstance(X, list):
            return X
        elif isinstance(X, np.ndarray):
            return X.tolist()
        elif isinstance(X, pd.DataFrame):
            return X.values.tolist()

    else:
        raise ValueError(
            f"Unsupported return type provided: {return_type}. Supported types "
            f"are 'df', 'array', and 'list'"
        )



from snpio.analysis.genotype_encoder import GenotypeEncoder
from snpio.filtering.nremover2 import NRemover2
from snpio.io.phylip_reader import PhylipReader
from snpio.io.structure_reader import StructureReader
from snpio.io.vcf_reader import VCFReader
from snpio.plotting.plotting import Plotting
